import textwrap
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

from lsp_tests.lsp_tutil import put_doc_with_cursor

try:
    from pygls.server import LanguageServer

    from debputy.lsp.lsp_debian_debputy_manifest import debputy_manifest_completer
    from debputy.lsp.debputy_ls import DebputyLanguageServer

    HAS_PYGLS = True
except ImportError:
    HAS_PYGLS = False


def _complete_at_req(uri: str, cursor_pos: types.Position) -> types.CompletionParams:
    return types.CompletionParams(types.TextDocumentIdentifier(uri), cursor_pos)


def test_basic_debputy_completer_empty(ls: "DebputyLanguageServer") -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label: m for m in completions}
    assert "definitions:" in keywords
    assert "manifest-version:" in keywords
    assert "installations:" in keywords

    expected_detail = "Define which paths are installed into which packages"

    assert keywords["installations:"].detail == expected_detail

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manif<CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "manifest-version:" in keywords
    assert "packages:" in keywords
    # We rely on client side filtering
    assert "installations:" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "definitions:" in keywords
    assert "installations:" in keywords
    assert "packages:" in keywords
    # Already completed
    assert "manifest-version:" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            sources:
              - foo
              - bar
        <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "definitions:" in keywords
    assert "packages:" in keywords
    # Already completed
    assert "manifest-version:" not in keywords
    assert "installations:" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            sources:
              - foo
              - bar
        packa<CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "definitions:" in keywords
    assert "packages:" in keywords
    # Already completed
    assert "manifest-version:" not in keywords
    assert "installations:" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            sources:
              - foo
              - bar
        packages:
          foo:
            services:
            - service: foo
              service-scope: user
        <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "definitions:" in keywords
    # Already completed
    assert "manifest-version:" not in keywords
    assert "installations:" not in keywords
    assert "packages:" not in keywords


def test_basic_debputy_completer_manifest_variable_value(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "'0.1'" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.<CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "'0.1'" in keywords


def test_basic_debputy_completer_install_rule_dispatch_key(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - <CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "install:" in keywords
    assert "install-doc:" in keywords
    assert "install-docs:" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - i<CURSOR>
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "install:" in keywords
    assert "install-doc:" in keywords
    assert "install-docs:" in keywords


def test_basic_debputy_completer_install_rule_install_keys(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install:
            <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "source:" in keywords
    assert "sources:" in keywords
    assert "as:" in keywords
    assert "dest-dir:" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install:
            sources:
            - foo
            - bar
            <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "dest-dir:" in keywords
    # Already completed
    assert "sources:" not in keywords

    # Not possible (conflict)
    assert "source:" not in keywords
    assert "as:" not in keywords


def test_basic_debputy_completer_packages_foo(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "binary-version:" in keywords
    assert "services:" in keywords
    assert "transformations:" in keywords


def test_basic_debputy_completer_packages_foo_xfail(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            bin<CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "binary-version:" in keywords
    assert "services:" in keywords
    assert "transformations:" in keywords


def test_basic_debputy_completer_services_service_scope_values(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            services:
            - service: foo
              service-scope: <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert keywords == {"system", "user"}

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            services:
            - service: foo
              service-scope: s<CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert keywords == {"system", "user"}

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            services:
            - service: foo
              service-scope: system
              enable-on-install: <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert keywords == {"true", "false"}

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        packages:
          foo:
            services:
            - service: foo
              service-scope: system
              enable-on-install: tr<CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    # "false" is ok, because we rely on client side filtering
    assert keywords == {"true", "false"}


def test_basic_debputy_completer_manifest_conditions(
    ls: "DebputyLanguageServer",
) -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            when: <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "cross-compiling" in keywords
    # Mapping-only forms are not applicable here
    assert "not" not in keywords
    assert "not:" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            when: c<CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "cross-compiling" in keywords
    # Mapping-only forms are not applicable here
    assert "not" not in keywords
    assert "not:" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            when:
              <CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "not:" in keywords
    # str-only forms are not applicable here
    assert "cross-compiling" not in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            when:
              n<CURSOR>
"""
        ),
    )
    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "not:" in keywords
    # str-only forms are not applicable here
    assert "cross-compiling" not in keywords


def test_basic_debputy_completer_mid_doc(ls: "DebputyLanguageServer") -> None:
    debputy_manifest_uri = "file:///nowhere/debian/debputy.manifest"
    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            s<CURSOR>
            - foo
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "sources:" in keywords

    cursor_pos = put_doc_with_cursor(
        ls,
        debputy_manifest_uri,
        "debian/debputy.manifest",
        textwrap.dedent(
            """\
        manifest-version: 0.1
        installations:
        - install-docs:
            s<CURSOR>:
            - foo
"""
        ),
    )

    completions = debputy_manifest_completer(
        ls,
        _complete_at_req(debputy_manifest_uri, cursor_pos),
    )
    assert isinstance(completions, list)
    keywords = {m.label for m in completions}
    assert "sources" in keywords



import textwrap

from debputy.plugin.api import (
    DebputyPluginInitializer,
    packager_provided_file_reference_documentation,
)
from debputy.plugin.debputy.metadata_detectors import (
    detect_systemd_tmpfiles,
    detect_kernel_modules,
    detect_icons,
    detect_gsettings_dependencies,
    detect_xfonts,
    detect_initramfs_hooks,
    detect_systemd_sysusers,
    detect_pycompile_files,
    translate_capabilities,
    pam_auth_update,
    auto_depends_arch_any_solink,
)
from debputy.plugin.debputy.paths import (
    SYSTEMD_TMPFILES_DIR,
    INITRAMFS_HOOK_DIR,
    GSETTINGS_SCHEMA_DIR,
    SYSTEMD_SYSUSERS_DIR,
)
from debputy.plugin.debputy.private_api import initialize_via_private_api


def initialize_debputy_features(api: DebputyPluginInitializer) -> None:
    initialize_via_private_api(api)
    declare_manifest_variables(api)
    register_packager_provided_files(api)
    register_package_metadata_detectors(api)


def declare_manifest_variables(api: DebputyPluginInitializer) -> None:
    api.manifest_variable(
        "path:BASH_COMPLETION_DIR",
        "/usr/share/bash-completion/completions",
        variable_reference_documentation="Directory to install bash completions into",
    )
    api.manifest_variable(
        "path:FISH_COMPLETION_DIR",
        "/usr/share/fish/vendor_completions.d",
        variable_reference_documentation="Directory to install fish completions into",
    )
    api.manifest_variable(
        "path:ZSH_COMPLETION_DIR",
        "/usr/share/zsh/vendor-completions",
        variable_reference_documentation="Directory to install zsh completions into",
    )
    api.manifest_variable(
        "path:GNU_INFO_DIR",
        "/usr/share/info",
        variable_reference_documentation="Directory to install GNU INFO files into",
    )

    api.manifest_variable(
        "token:NL",
        "\n",
        variable_reference_documentation="Literal newline (linefeed) character",
    )
    api.manifest_variable(
        "token:NEWLINE",
        "\n",
        variable_reference_documentation="Literal newline (linefeed) character",
    )
    api.manifest_variable(
        "token:TAB",
        "\t",
        variable_reference_documentation="Literal tab character",
    )
    api.manifest_variable(
        "token:OPEN_CURLY_BRACE",
        "{",
        variable_reference_documentation='Literal "{" character',
    )
    api.manifest_variable(
        "token:CLOSE_CURLY_BRACE",
        "}",
        variable_reference_documentation='Literal "}" character',
    )
    api.manifest_variable(
        "token:DOUBLE_OPEN_CURLY_BRACE",
        "{{",
        variable_reference_documentation='Literal "{{" character - useful to avoid triggering a substitution',
    )
    api.manifest_variable(
        "token:DOUBLE_CLOSE_CURLY_BRACE",
        "}}",
        variable_reference_documentation='Literal "}}" string - useful to avoid triggering a substitution',
    )


def register_package_metadata_detectors(api: DebputyPluginInitializer) -> None:
    api.metadata_or_maintscript_detector("systemd-tmpfiles", detect_systemd_tmpfiles)
    api.metadata_or_maintscript_detector("systemd-sysusers", detect_systemd_sysusers)
    api.metadata_or_maintscript_detector("kernel-modules", detect_kernel_modules)
    api.metadata_or_maintscript_detector("icon-cache", detect_icons)
    api.metadata_or_maintscript_detector(
        "gsettings-dependencies",
        detect_gsettings_dependencies,
    )
    api.metadata_or_maintscript_detector("xfonts", detect_xfonts)
    api.metadata_or_maintscript_detector("initramfs-hooks", detect_initramfs_hooks)
    api.metadata_or_maintscript_detector("pycompile-files", detect_pycompile_files)
    api.metadata_or_maintscript_detector(
        "translate-capabilities",
        translate_capabilities,
    )
    api.metadata_or_maintscript_detector("pam-auth-update", pam_auth_update)
    api.metadata_or_maintscript_detector(
        "auto-depends-arch-any-solink",
        auto_depends_arch_any_solink,
    )


def register_packager_provided_files(api: DebputyPluginInitializer) -> None:
    api.packager_provided_file(
        "tmpfiles",
        f"{SYSTEMD_TMPFILES_DIR}/{{name}}.conf",
        reference_documentation=packager_provided_file_reference_documentation(
            format_documentation_uris=["man:tmpfiles.d(5)"]
        ),
    )
    api.packager_provided_file(
        "sysusers",
        f"{SYSTEMD_SYSUSERS_DIR}/{{name}}.conf",
        reference_documentation=packager_provided_file_reference_documentation(
            format_documentation_uris=["man:sysusers.d(5)"]
        ),
    )
    api.packager_provided_file(
        "bash-completion", "/usr/share/bash-completion/completions/{name}"
    )
    api.packager_provided_file(
        "fish-completion", "/usr/share/fish/vendor_completions.d/{name}"
    )
    api.packager_provided_file(
        "zsh-completion", "/usr/share/zsh/vendor-completions/{name}"
    )
    api.packager_provided_file(
        "bug-script",
        "./usr/share/bug/{name}/script",
        default_mode=0o0755,
        allow_name_segment=False,
    )
    api.packager_provided_file(
        "bug-control",
        "/usr/share/bug/{name}/control",
        allow_name_segment=False,
    )

    api.packager_provided_file(
        "bug-presubj",
        "/usr/share/bug/{name}/presubj",
        allow_name_segment=False,
    )

    api.packager_provided_file("pam", "/usr/lib/pam.d/{name}")
    api.packager_provided_file(
        "ppp.ip-up",
        "/etc/ppp/ip-up.d/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "ppp.ip-down",
        "/etc/ppp/ip-down.d/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "lintian-overrides",
        "/usr/share/lintian/overrides/{name}",
        allow_name_segment=False,
    )
    api.packager_provided_file("logrotate", "/etc/logrotate.d/{name}")
    api.packager_provided_file(
        "logcheck.cracking",
        "/etc/logcheck/cracking.d/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    api.packager_provided_file(
        "logcheck.violations",
        "/etc/logcheck/violations.d/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    api.packager_provided_file(
        "logcheck.violations.ignore",
        "/etc/logcheck/violations.ignore.d/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    api.packager_provided_file(
        "logcheck.ignore.workstation",
        "/etc/logcheck/ignore.d.workstation/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    api.packager_provided_file(
        "logcheck.ignore.server",
        "/etc/logcheck/ignore.d.server/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )
    api.packager_provided_file(
        "logcheck.ignore.paranoid",
        "/etc/logcheck/ignore.d.paranoid/{name}",
        post_formatting_rewrite=_replace_dot_with_underscore,
    )

    api.packager_provided_file("mime", "/usr/lib/mime/packages/{name}")
    api.packager_provided_file("sharedmimeinfo", "/usr/share/mime/packages/{name}.xml")

    api.packager_provided_file(
        "if-pre-up",
        "/etc/network/if-pre-up.d/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "if-up",
        "/etc/network/if-up.d/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "if-down",
        "/etc/network/if-down.d/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "if-post-down",
        "/etc/network/if-post-down.d/{name}",
        default_mode=0o0755,
    )

    api.packager_provided_file(
        "cron.hourly",
        "/etc/cron.hourly/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "cron.daily",
        "/etc/cron.daily/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "cron.weekly",
        "/etc/cron.weekly/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "cron.monthly",
        "./etc/cron.monthly/{name}",
        default_mode=0o0755,
    )
    api.packager_provided_file(
        "cron.yearly",
        "/etc/cron.yearly/{name}",
        default_mode=0o0755,
    )
    # cron.d uses 0644 unlike the others
    api.packager_provided_file(
        "cron.d",
        "/etc/cron.d/{name}",
        reference_documentation=packager_provided_file_reference_documentation(
            format_documentation_uris=["man:crontab(5)"]
        ),
    )

    api.packager_provided_file(
        "initramfs-hook", f"{INITRAMFS_HOOK_DIR}/{{name}}", default_mode=0o0755
    )

    api.packager_provided_file("modprobe", "/etc/modprobe.d/{name}.conf")

    api.packager_provided_file(
        "init",
        "/etc/init.d/{name}",
        default_mode=0o755,
    )
    api.packager_provided_file("default", "/etc/default/{name}")

    for stem in [
        "mount",
        "path",
        "service",
        "socket",
        "target",
        "timer",
    ]:
        api.packager_provided_file(
            stem,
            f"/usr/lib/systemd/system/{{name}}.{stem}",
            reference_documentation=packager_provided_file_reference_documentation(
                format_documentation_uris=[f"man:systemd.{stem}(5)"]
            ),
        )

    for stem in [
        "path",
        "service",
        "socket",
        "target",
        "timer",
    ]:
        api.packager_provided_file(
            f"@{stem}", f"/usr/lib/systemd/system/{{name}}@.{stem}"
        )

    # api.packager_provided_file(
    #     "udev",
    #     "./lib/udev/rules.d/{priority:02}-{name}.rules",
    #     default_priority=60,
    # )

    api.packager_provided_file(
        "gsettings-override",
        f"{GSETTINGS_SCHEMA_DIR}/{{priority:02}}_{{name}}.gschema.override",
        default_priority=10,
    )

    # Special-cases that will probably not be a good example for other plugins
    api.packager_provided_file(
        "changelog",
        # The "changelog.Debian" gets renamed to "changelog" for native packages elsewhere.
        # Also, the changelog trimming is also done elsewhere.
        "/usr/share/doc/{name}/changelog.Debian",
        allow_name_segment=False,
        packageless_is_fallback_for_all_packages=True,
        reference_documentation=packager_provided_file_reference_documentation(
            description=textwrap.dedent(
                """\
                This file is the changelog of the package and is mandatory.

                The changelog contains the version of the source package and is mandatory for all
                packages.

                Use `dch --create` to create the changelog.

                In theory, the binary package can have a different changelog than the source
                package (by having `debian/binary-package.changelog`). However, it is generally
                not useful and leads to double administration. It has not been used in practice.
            """
            ),
            format_documentation_uris=[
                "man:deb-changelog(5)",
                "https://www.debian.org/doc/debian-policy/ch-source.html#debian-changelog-debian-changelog",
                "man:dch(1)",
            ],
        ),
    )
    api.packager_provided_file(
        "copyright",
        "/usr/share/doc/{name}/copyright",
        allow_name_segment=False,
        packageless_is_fallback_for_all_packages=True,
        reference_documentation=packager_provided_file_reference_documentation(
            description=textwrap.dedent(
                """\
                This file documents the license and copyright information of the binary package.
                Packages aimed at the Debian archive (and must derivatives thereof) must have this file.

                For packages not aimed at Debian, the file can still be useful to convey the license
                terms of the package (which is often a requirement in many licenses). However, it is
                not a strict *technical* requirement. Whether it is a legal requirement depends on
                license.

                Often, the same file can be used for all packages. In the extremely rare case where
                one binary package has a "vastly different" license than the other packages, you can
                provide a package specific version for that package.
            """
            ),
            format_documentation_uris=[
                "https://www.debian.org/doc/debian-policy/ch-source.html#copyright-debian-copyright",
                "https://www.debian.org/doc/debian-policy/ch-docs.html#s-copyrightfile",
                "https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/",
            ],
        ),
    )
    api.packager_provided_file(
        "NEWS",
        "/usr/share/doc/{name}/NEWS.Debian",
        allow_name_segment=False,
        packageless_is_fallback_for_all_packages=True,
        reference_documentation=packager_provided_file_reference_documentation(
            description=textwrap.dedent(
                """\
            Important news that should be shown to the user/admin when upgrading. If a system has
            apt-listchanges installed, then contents of this file will be shown prior to upgrading
            the package.

            Uses a similar format to that of debian/changelog (create with `dch --news --create`).
            """
            ),
            format_documentation_uris=[
                "https://www.debian.org/doc/manuals/developers-reference/best-pkging-practices.en.html#supplementing-changelogs-with-news-debian-files",
                "man:dch(1)",
            ],
        ),
    )
    api.packager_provided_file(
        "README.Debian",
        "/usr/share/doc/{name}/README.Debian",
        allow_name_segment=False,
    )
    api.packager_provided_file(
        "TODO",
        "/usr/share/doc/{name}/TODO.Debian",
        allow_name_segment=False,
    )
    # From dh-python / dh_python3
    # api.packager_provided_file(
    #     "bcep",
    #     "/usr/share/python3/bcep/{name}",
    #     allow_name_segment=False,
    # )


def _replace_dot_with_underscore(x: str) -> str:
    return x.replace(".", "_")



try:
    from typing import TypeVar, Union, Tuple, List, Callable, Iterator, TYPE_CHECKING

    if TYPE_CHECKING:
        from .tokens import Deb822Token, Deb822FieldNameToken
        from .parsing import (
            Deb822Element,
            Deb822CommentElement,
            Deb822ParsedValueElement,
        )
        from .formatter import FormatterContentToken

    TokenOrElement = Union["Deb822Element", "Deb822Token"]
    TE = TypeVar("TE", bound=TokenOrElement)

    # Used as a resulting element for "mapping" functions that map TE -> R (see _combine_parts)
    R = TypeVar("R", bound="Deb822Element")

    VE = TypeVar("VE", bound="Deb822Element")

    ST = TypeVar("ST", bound="Deb822Token")

    # Internal type for part of the paragraph key.  Used to facility _unpack_key.
    ParagraphKeyBase = Union["Deb822FieldNameToken", str]

    ParagraphKey = Union[ParagraphKeyBase, Tuple[str, int]]

    Commentish = Union[List[str], "Deb822CommentElement"]

    FormatterCallback = Callable[
        [str, "FormatterContentToken", Iterator["FormatterContentToken"]],
        Iterator[Union["FormatterContentToken", str]],
    ]
    try:
        # Set __doc__ attributes if possible
        TE.__doc__ = """
        Generic "Token or Element" type
        """
        R.__doc__ = """
        For internal usage in _deb822_repro
        """
        VE.__doc__ = """
        Value type/element in a list interpretation of a field value
        """
        ST.__doc__ = """
        Separator type/token in a list interpretation of a field value
        """
        ParagraphKeyBase.__doc__ = """
        For internal usage in _deb822_repro
        """
        ParagraphKey.__doc__ = """
        Anything accepted as a key for a paragraph field lookup.  The simple case being
        a str. Alternative variants are mostly interesting for paragraphs with repeated
        fields (to enable unambiguous lookups)
        """
        Commentish.__doc__ = """
        Anything accepted as input for a Comment. The simple case is the list
        of string (each element being a line of comment). The alternative format is
        there for enable reuse of an existing element (e.g. to avoid "unpacking"
        only to "re-pack" an existing comment element).
        """
        FormatterCallback.__doc__ = """\
        Formatter callback used with the round-trip safe parser

        See debian._repro_deb822.formatter.format_field for details
        """
    except AttributeError:
        # Python 3.5 does not allow update to the __doc__ attribute - ignore that
        pass
except ImportError:
    pass


class AmbiguousDeb822FieldKeyError(KeyError):
    """Specialized version of KeyError to denote a valid but ambiguous field name

    This exception occurs if:
      * the field is accessed via a str on a configured view that does not automatically
        resolve ambiguous field names (see Deb822ParagraphElement.configured_view), AND
      * a concrete paragraph contents a repeated field (which is not valid in deb822
        but the module supports parsing them)

    Note that the default is to automatically resolve ambiguous fields. Accordingly
    you will only see this exception if you have "opted in" on wanting to know that
    the lookup was ambiguous.

    The ambiguity can be resolved by using a tuple of (<field-name>, <filed-index>)
    instead of <field-name>.
    """


class SyntaxOrParseError(ValueError):
    """Specialized version of ValueError for syntax/parse errors."""



from .attributes import Style as Style
from .background import Back as Back
from .colored import Colored as Colored, attr as attr, back as back, back_rgb as back_rgb, bg as bg, fg as fg, fore as fore, fore_rgb as fore_rgb, set_tty_aware as set_tty_aware, style as style, stylize as stylize, stylize_interactive as stylize_interactive
from .controls import Controls as Controls
from .cprint import cprint as cprint
from .foreground import Fore as Fore

__version__: str



import pytest

from debputy.manifest_parser.base_types import SymbolicMode


@pytest.mark.parametrize(
    "base_mode,is_dir,symbolic_mode,expected",
    [
        (0o0644, False, "u+rwX,og=rX", 0o0644),
        (0o0000, False, "u+rwX,og=rX", 0o0644),
        (0o0400, True, "u+rwX,og=rX", 0o0755),
        (0o0000, True, "u+rwX,og=rX", 0o0755),
        (0o2400, False, "u+rwxs,og=rx", 0o04755),
        (0o7400, False, "u=rwX,og=rX", 0o0644),
        (0o0641, False, "u=rwX,og=rX", 0o0755),
        (0o4755, False, "a-x", 0o04644),
    ],
)
def test_generate_deb_filename(
    attribute_path, base_mode, is_dir, symbolic_mode, expected
):
    print(attribute_path.path)
    parsed_mode = SymbolicMode.parse_filesystem_mode(symbolic_mode, attribute_path)
    actual = parsed_mode.compute_mode(base_mode, is_dir)
    assert oct(actual)[2:] == oct(expected)[2:]



import contextlib
import dataclasses
import os.path
import subprocess
from typing import (
    Optional,
    FrozenSet,
    final,
    TYPE_CHECKING,
    Union,
    Annotated,
    List,
    NotRequired,
    Literal,
    Any,
    Type,
    TypeVar,
    Self,
    Sequence,
    Callable,
    Container,
    Iterable,
    is_typeddict,
)

from debputy.exceptions import PluginAPIViolationError, PluginInitializationError
from debputy.manifest_conditions import ManifestCondition
from debputy.manifest_parser.base_types import (
    BuildEnvironmentDefinition,
    DebputyParsedContentStandardConditional,
    FileSystemExactMatchRule,
)
from debputy.manifest_parser.exceptions import (
    ManifestParseException,
    ManifestInvalidUserDataException,
)
from debputy.manifest_parser.parse_hints import DebputyParseHint
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.tagging_types import DebputyDispatchableType
from debputy.manifest_parser.util import AttributePath
from debputy.packages import BinaryPackage
from debputy.plugin.api.spec import (
    ParserDocumentation,
    DebputyIntegrationMode,
    BuildSystemManifestRuleMetadata,
    _DEBPUTY_DISPATCH_METADATA_ATTR_NAME,
    VirtualPath,
)
from debputy.plugin.plugin_state import run_in_context_of_plugin
from debputy.substitution import Substitution
from debputy.types import EnvironmentModification
from debputy.util import (
    run_build_system_command,
    _debug_log,
    _info,
    _warn,
    generated_root_directory,
)

if TYPE_CHECKING:
    from debputy.build_support.build_context import BuildContext
    from debputy.highlevel_manifest import HighLevelManifest


AT = TypeVar("AT")
BSR = TypeVar("BSR", bound="BuildSystemRule")
BSPF = TypeVar("BSPF", bound="BuildRuleDefinitionBase")


@dataclasses.dataclass(slots=True, frozen=True)
class BuildSystemCharacteristics:
    out_of_source_builds: Literal[
        "required",
        "supported-and-default",
        "supported-but-not-default",
        "not-supported",
    ]


class CleanHelper:
    def schedule_removal_of_files(self, *args: str) -> None:
        """Schedule removal of these files

        This will remove the provided files in bulk. The files are not guaranteed
        to be deleted in any particular order. If anything needs urgent removal,
        `os.unlink` can be used directly.

        Note: Symlinks will **not** be followed. If a symlink and target must
        be deleted, ensure both are passed.


        :param args: Path names to remove. Each must be removable with
          `os.unlink`
        """
        raise NotImplementedError

    def schedule_removal_of_directories(self, *args: str) -> None:
        """Schedule removal of these directories

        This will remove the provided dirs in bulk. The dirs are not guaranteed
        to be deleted in any particular order. If anything needs urgent removal,
        then it can be done directly instead of passing it to this method.

        If anything needs urgent removal, then it can be removed immediately.

        :param args: Path names to remove.
        """
        raise NotImplementedError


class BuildRuleParsedFormat(DebputyParsedContentStandardConditional):
    name: NotRequired[str]
    for_packages: NotRequired[
        Annotated[
            Union[BinaryPackage, List[BinaryPackage]],
            DebputyParseHint.manifest_attribute("for"),
        ]
    ]
    environment: NotRequired[BuildEnvironmentDefinition]


class OptionalBuildDirectory(BuildRuleParsedFormat):
    build_directory: NotRequired[FileSystemExactMatchRule]


class OptionalInSourceBuild(BuildRuleParsedFormat):
    perform_in_source_build: NotRequired[bool]


class OptionalInstallDirectly(BuildRuleParsedFormat):
    install_directly_to_package: NotRequired[bool]


BuildSystemDefinition = Union[
    BuildRuleParsedFormat,
    OptionalBuildDirectory,
    OptionalInSourceBuild,
    OptionalInstallDirectly,
]


class BuildRule(DebputyDispatchableType):
    __slots__ = (
        "_auto_generated_stem",
        "_name",
        "_for_packages",
        "_manifest_condition",
        "_attribute_path",
        "_environment",
        "_substitution",
    )

    def __init__(
        self,
        attributes: BuildRuleParsedFormat,
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__()

        self._name = attributes.get("name")
        for_packages = attributes.get("for_packages")

        if for_packages is None:
            if isinstance(parser_context, ParserContextData):
                all_binaries = parser_context.binary_packages.values()
            else:
                all_binaries = parser_context.all_packages
            self._for_packages = frozenset(all_binaries)
        else:
            self._for_packages = frozenset(
                for_packages if isinstance(for_packages, list) else [for_packages]
            )
        self._manifest_condition = attributes.get("when")
        self._attribute_path = attribute_path
        self._substitution = parser_context.substitution
        self._auto_generated_stem: Optional[str] = None
        environment = attributes.get("environment")
        if environment is None:
            assert isinstance(parser_context, ParserContextData)
            self._environment = parser_context.resolve_build_environment(
                None,
                attribute_path,
            )
        else:
            self._environment = environment

    @final
    @property
    def name(self) -> Optional[str]:
        return self._name

    @final
    @property
    def attribute_path(self) -> AttributePath:
        return self._attribute_path

    @final
    @property
    def manifest_condition(self) -> Optional[ManifestCondition]:
        return self._manifest_condition

    @final
    @property
    def for_packages(self) -> FrozenSet[BinaryPackage]:
        return self._for_packages

    @final
    @property
    def substitution(self) -> Substitution:
        return self._substitution

    @final
    @property
    def environment(self) -> BuildEnvironmentDefinition:
        return self._environment

    @final
    @property
    def auto_generated_stem(self) -> str:
        stem = self._auto_generated_stem
        if stem is None:
            raise AssertionError(
                "The auto-generated-stem is not available at this time"
            )
        return stem

    @final
    @auto_generated_stem.setter
    def auto_generated_stem(self, value: str) -> None:
        if self._auto_generated_stem is not None:
            raise AssertionError("The auto-generated-stem should only be set once")
        assert value is not None
        self._auto_generated_stem = value

    @final
    def run_build(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        run_in_context_of_plugin(
            self._debputy_plugin,
            self.perform_build,
            context,
            manifest,
            **kwargs,
        )

    def perform_build(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        raise NotImplementedError

    @property
    def is_buildsystem(self) -> bool:
        return False

    @property
    def name_or_tag(self) -> str:
        name = self.name
        if name is None:
            return self.auto_generated_stem
        return name

    def install_dest_dir(self) -> Optional[str]:
        return None


def _is_type_or_none(v: Optional[Any], expected_type: Type[AT]) -> Optional[AT]:
    if isinstance(v, expected_type):
        return v
    return None


class BuildSystemRule(BuildRule):

    __slots__ = (
        "_build_directory",
        "source_directory",
        "install_directly_to_package",
        "perform_in_source_build",
    )

    def __init__(
        self,
        attributes: BuildSystemDefinition,
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(attributes, attribute_path, parser_context)
        build_directory = _is_type_or_none(
            attributes.get("build_directory"), FileSystemExactMatchRule
        )
        if build_directory is not None:
            self._build_directory = build_directory.match_rule.path
        else:
            self._build_directory = None
        self.source_directory = "."
        self.install_directly_to_package = False
        self.perform_in_source_build = _is_type_or_none(
            attributes.get("perform_in_source_build"), bool
        )
        install_directly_to_package = _is_type_or_none(
            attributes.get("install_directly_to_package"), bool
        )
        if install_directly_to_package is None:
            self.install_directly_to_package = len(self.for_packages) == 1
        elif install_directly_to_package and len(self.for_packages) > 1:
            idtp_path = attribute_path["install_directly_to_package"].path
            raise ManifestParseException(
                f'The attribute "install-directly-to-package" ({idtp_path}) cannot'
                " be true when the build system applies to multiple packages."
            )
        else:
            self.install_directly_to_package = install_directly_to_package

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        """Check if the build system apply automatically.

        This class method is called when the manifest does not declare any build rules at
        all.

        :param source_root: The source root (the directory containing `debian/`). Usually,
          the detection code would look at this for files related to the upstream build system.
        :param args: For future compat, new arguments might appear as positional arguments.
        :param kwargs: For future compat, new arguments might appear as keyword argument.
        :return: True if the build system can be used, False when it would not be useful
          to use the build system (at least with all defaults).
          Note: Be sure to use proper `bool` return values. The calling code does an
          `isinstance` check to ensure that the version of `debputy` supports the
          auto-detector (in case the return type is ever expanded in the future).
        """
        return False

    @property
    def out_of_source_build(self) -> bool:
        build_directory = self.build_directory
        return build_directory != self.source_directory

    @property
    def build_directory(self) -> str:
        directory = self._build_directory
        if directory is None:
            return self.source_directory
        return directory

    @contextlib.contextmanager
    def dump_logs_on_error(self, *logs: str) -> None:
        """Context manager that will dump logs to stdout on error

        :param logs: The logs to be dumped. Relative path names are assumed to be relative to
          the build directory.
        """
        try:
            yield
        except (Exception, KeyboardInterrupt, SystemExit):
            _warn(
                "Error occurred, attempting to provide relevant logs as requested by the build system provider"
            )
            found_any = False
            for log in logs:
                if not os.path.isabs(log):
                    log = self.build_dir_path(log)
                if not os.path.isfile(log):
                    _info(
                        f'Would have pushed "{log}" to stdout, but it does not exist.'
                    )
                    continue
                subprocess.run(["tail", "-v", "-n", "+0", log])
                found_any = True
            if not found_any:
                _warn(
                    f"None of the logs provided were available (relative to build directory): {', '.join(logs)}"
                )
            raise

    @final
    def run_clean(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: CleanHelper,
        **kwargs,
    ) -> None:
        run_in_context_of_plugin(
            self._debputy_plugin,
            self.perform_clean,
            context,
            manifest,
            clean_helper,
            **kwargs,
        )

    def perform_clean(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: CleanHelper,
        **kwargs,
    ) -> None:
        raise NotImplementedError

    def ensure_build_dir_exists(self) -> None:
        build_dir = self.build_directory
        source_dir = self.source_directory
        if build_dir == source_dir:
            return
        os.makedirs(build_dir, mode=0o755, exist_ok=True)

    def build_dir_path(self, /, path: str = "") -> str:
        build_dir = self.build_directory
        if path == "":
            return build_dir
        return os.path.join(build_dir, path)

    def relative_from_builddir_to_source(
        self,
        path_in_source_dir: Optional[str] = None,
    ) -> str:
        build_dir = self.build_directory
        source_dir = self.source_directory

        if path_in_source_dir is None:
            source_path = source_dir
        else:
            source_path = os.path.join(source_dir, path_in_source_dir)

        if build_dir == source_dir:
            return source_path

        return os.path.relpath(source_path, build_dir)

    @final
    @property
    def is_buildsystem(self) -> bool:
        return True


class StepBasedBuildSystemRule(BuildSystemRule):

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        raise NotImplementedError

    @final
    def perform_clean(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: CleanHelper,
        **kwargs,
    ) -> None:
        self._check_characteristics()
        self.before_first_impl_step(stage="clean")
        self.clean_impl(context, manifest, clean_helper, **kwargs)
        if self.out_of_source_build:
            build_directory = self.build_directory
            assert build_directory is not None
            if os.path.lexists(build_directory):
                clean_helper.schedule_removal_of_directories(build_directory)
        dest_dir = self.resolve_dest_dir()
        if not isinstance(dest_dir, BinaryPackage):
            clean_helper.schedule_removal_of_directories(dest_dir)

    @final
    def install_dest_dir(self) -> Optional[str]:
        dest_dir = self.resolve_dest_dir()
        if isinstance(dest_dir, BinaryPackage):
            return None
        return dest_dir

    @final
    def perform_build(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._check_characteristics()
        self.before_first_impl_step(stage="build")
        self.configure_impl(context, manifest, **kwargs)
        self.build_impl(context, manifest, **kwargs)
        if context.should_run_tests:
            self.test_impl(context, manifest, **kwargs)
        dest_dir = self.resolve_dest_dir()
        if isinstance(dest_dir, BinaryPackage):
            dest_dir = f"debian/{dest_dir.name}"
        # Make it absolute for everyone (that worked for debhelper).
        # At least autoconf's "make install" requires an absolute path, so making is
        # relative would have at least one known issue.
        abs_dest_dir = os.path.abspath(dest_dir)
        self.install_impl(context, manifest, abs_dest_dir, **kwargs)
        if os.path.isdir(abs_dest_dir):
            generated_root_directory(abs_dest_dir)

    def before_first_impl_step(
        self,
        /,
        stage: Literal["build", "clean"],
        **kwargs,
    ) -> None:
        """Called before any `*_impl` method is called.

        This can be used to validate input against data that is not available statically
        (that is, it will be checked during build but not in static checks). An example
        is that the `debhelper` build system uses this to validate the provided `dh-build-system`
        to ensure that `debhelper` knows about the build system. This check cannot be done
        statically since the build system is only required to be available in a chroot build
        and not on the host system.

        The method can also be used to compute common state for all the `*_impl` methods that
        is awkward to do in `__init__`. Note there is no data sharing between the different
        stages. This has to do with how `debputy` will be called (usually `clean` followed by
        a source package assembly in `dpkg` and then `build`).

        The check is done both on build and on clean before the relevant implementation methods
        are invoked.

        Any exception will abort the build. Prefer to raise ManifestInvalidUserDataException
        exceptions for issues related to incorrect data.

        The method is not invoked if the steps are skipped, which can happen with build profiles
        or arch:any vs. arch:all builds.

        :param stage: A discriminator variable to determine which kind of steps will be invoked
        after this method returns. For state initialization, this can be useful if the state
        is somewhat expensive and not needed for `clean`.
        """
        pass

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        """Called to handle the "configure" and "build" part of the build

        This is basically a mix of `dh_auto_configure` and `dh_auto_build` from `debhelper`.
        If the upstream build also runs test as a part of the build, this method should
        check `context.should_run_tests` and pass the relevant flags to disable tests when
        `context.should_run_tests` is false.
        """
        raise NotImplementedError

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        """Called to handle the "configure" and "build" part of the build

        This is basically a mix of `dh_auto_configure` and `dh_auto_build` from `debhelper`.
        If the upstream build also runs test as a part of the build, this method should
        check `context.should_run_tests` and pass the relevant flags to disable tests when
        `context.should_run_tests` is false.
        """
        raise NotImplementedError

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        """Called to handle the "test" part of the build

        This is basically `dh_auto_test` from `debhelper`.

        Note: This will be skipped when `context.should_run_tests` is False. Therefore, the
        method can assume that when invoked then tests must be run.

        It is always run after `configure_and_build_impl`.
        """
        raise NotImplementedError

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        """Called to handle the "install" part of the build

        This is basically `dh_auto_install` from `debhelper`.

        The `dest_dir` attribute is what the upstream should install its data into. It
        follows the `DESTDIR` convention from autoconf/make. The `dest_dir` should not
        be second-guessed since `debputy` will provide automatically as a search path
        for installation rules when relevant.

        It is always run after `configure_and_build_impl` and, if relevant, `test_impl`.
        """
        raise NotImplementedError

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        """Called to handle the "clean" part of the build

        This is basically `dh_auto_clean` from `debhelper`.

        For out-of-source builds, `debputy` will remove the build directory for you
        if it exists (when this method returns). This method is only "in-source" cleaning
        or for "dirty" state left outside the designated build directory.

        Note that state *cannot* be shared between `clean` and other steps due to limitations
        of how the Debian build system works in general.
        """
        raise NotImplementedError

    def _check_characteristics(self) -> None:
        characteristics = self.characteristics()

        _debug_log(f"Characteristics for {self.name_or_tag} {self.__class__.__name__} ")

        if self.out_of_source_build and self.perform_in_source_build:
            raise ManifestInvalidUserDataException(
                f"Cannot use 'build-directory' with 'perform-in-source-build' at {self.attribute_path.path}"
            )
        if (
            characteristics.out_of_source_builds == "required"
            and self.perform_in_source_build
        ):
            path = self.attribute_path["perform_in_source_build"].path_key_lc

            # FIXME: How do I determine the faulty plugin from here.
            raise PluginAPIViolationError(
                f"The build system {self.__class__.__qualname__} had an perform-in-source-build attribute, but claims"
                f" it requires out of source builds. Please file a bug against the provider asking them not to use"
                f' "{OptionalInSourceBuild.__name__}" as base for their build system definition or tweak'
                f" the characteristics of the build system as the current combination is inconsistent."
                f" The offending definition is at {path}."
            )

        if (
            characteristics.out_of_source_builds
            in ("required", "supported-and-default")
            and not self.out_of_source_build
        ):

            if not self.perform_in_source_build:
                self._build_directory = self._pick_build_dir()
            else:
                assert characteristics.out_of_source_builds != "required"
        elif (
            characteristics.out_of_source_builds == "not-supported"
            and self.out_of_source_build
        ):
            path = self.attribute_path["build_directory"].path_key_lc

            # FIXME: How do I determine the faulty plugin from here.
            raise PluginAPIViolationError(
                f"The build system {self.__class__.__qualname__} had a build-directory attribute, but claims it does"
                f" not support out of source builds. Please file a bug against the provider asking them not to use"
                f' "{OptionalBuildDirectory.__name__}" as base for their build system definition or tweak'
                f" the characteristics of the build system as the current combination is inconsistent."
                f" The offending definition is at {path}."
            )

    def _pick_build_dir(self) -> str:
        tag = self.name if self.name is not None else self.auto_generated_stem
        if tag == "":
            return "_build"
        return f"_build-{tag}"

    @final
    def resolve_dest_dir(self) -> Union[str, BinaryPackage]:
        auto_generated_stem = self.auto_generated_stem
        if self.install_directly_to_package:
            assert len(self.for_packages) == 1
            return next(iter(self.for_packages))
        if auto_generated_stem == "":
            return "debian/tmp"
        return f"debian/tmp-{auto_generated_stem}"


# Using the same logic as debhelper for the same reasons.
def _make_target_exists(make_cmd: str, target: str, *, directory: str = ".") -> bool:
    cmd = [
        make_cmd,
        "-s",
        "-n",
        "--no-print-directory",
    ]
    if directory and directory != ".":
        cmd.append("-C")
        cmd.append(directory)
    cmd.append(target)
    env = dict(os.environ)
    env["LC_ALL"] = "C.UTF-8"
    try:
        res = subprocess.run(
            cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            env=env,
            restore_signals=True,
        )
    except FileNotFoundError:
        return False

    options = (
        f"*** No rule to make target '{target}",
        f"*** No rule to make target `{target}",
    )

    stdout = res.stdout.decode("utf-8")
    return not any(o in stdout for o in options)


def _find_first_existing_make_target(
    make_cmd: str,
    targets: Sequence[str],
    *,
    directory: str = ".",
) -> Optional[str]:
    for target in targets:
        if _make_target_exists(make_cmd, target, directory=directory):
            return target
    return None


_UNSET = object()


class NinjaBuildSupport:
    __slots__ = ("_provided_ninja_program", "_build_system_rule")

    def __init__(
        self,
        provided_ninja_program: str,
        build_system_rule: BuildSystemRule,
    ) -> None:
        self._provided_ninja_program = provided_ninja_program
        self._build_system_rule = build_system_rule

    @classmethod
    def from_build_system(
        cls,
        build_system: BuildSystemRule,
        *,
        ninja_program: Optional[str] = None,
    ) -> Self:
        if ninja_program is None:
            ninja_program = "ninja"
        return cls(ninja_program, build_system)

    @property
    def _directory(self) -> str:
        return self._build_system_rule.build_directory

    def _pick_directory(
        self, arg: Union[Optional[str], _UNSET] = _UNSET
    ) -> Optional[str]:
        if arg is _UNSET:
            return self._directory
        return arg

    def run_ninja_build(
        self,
        build_context: "BuildContext",
        *ninja_args: str,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
        enable_parallelization: bool = True,
    ) -> None:
        extra_ninja_args = []
        if not build_context.is_terse_build:
            extra_ninja_args.append("-v")
        self._run_ninja(
            build_context,
            *extra_ninja_args,
            *ninja_args,
            env_mod=env_mod,
            directory=directory,
            enable_parallelization=enable_parallelization,
        )

    def run_ninja_test(
        self,
        build_context: "BuildContext",
        *ninja_args: str,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
        enable_parallelization: bool = True,
    ) -> None:
        self._run_ninja(
            build_context,
            "test",
            *ninja_args,
            env_mod=env_mod,
            directory=directory,
            enable_parallelization=enable_parallelization,
        )

    def run_ninja_install(
        self,
        build_context: "BuildContext",
        dest_dir: str,
        *ninja_args: str,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
        # debhelper never had parallel installs, so we do not have it either for now.
        enable_parallelization: bool = False,
    ) -> None:
        install_env_mod = EnvironmentModification(replacements=(("DESTDIR", dest_dir),))
        if env_mod is not None:
            install_env_mod = install_env_mod.combine(env_mod)
        self._run_ninja(
            build_context,
            "install",
            *ninja_args,
            directory=directory,
            env_mod=install_env_mod,
            enable_parallelization=enable_parallelization,
        )

    def run_ninja_clean(
        self,
        build_context: "BuildContext",
        *ninja_args: str,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
        enable_parallelization: bool = True,
    ) -> None:
        self._run_ninja(
            build_context,
            "clean",
            *ninja_args,
            env_mod=env_mod,
            directory=directory,
            enable_parallelization=enable_parallelization,
        )

    def _run_ninja(
        self,
        build_context: "BuildContext",
        *ninja_args: str,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
        enable_parallelization: bool = True,
    ) -> None:
        extra_ninja_args = []
        limit = (
            build_context.parallelization_limit(support_zero_as_unlimited=True)
            if enable_parallelization
            else 1
        )
        extra_ninja_args.append(f"-j{limit}")
        ninja_env_mod = EnvironmentModification(replacements=(("LC_ALL", "C.UTF-8"),))
        if env_mod is not None:
            ninja_env_mod = ninja_env_mod.combine(env_mod)
        run_build_system_command(
            self._provided_ninja_program,
            *extra_ninja_args,
            *ninja_args,
            cwd=self._pick_directory(directory),
            env_mod=ninja_env_mod,
        )


class MakefileSupport:

    __slots__ = ("_provided_make_program", "_build_system_rule")

    def __init__(
        self,
        make_program: str,
        build_system_rule: BuildSystemRule,
    ) -> None:
        self._provided_make_program = make_program
        self._build_system_rule = build_system_rule

    @classmethod
    def from_build_system(
        cls,
        build_system: BuildSystemRule,
        *,
        make_program: Optional[str] = None,
    ) -> Self:
        if make_program is None:
            make_program = os.environ.get("MAKE", "make")
        return cls(make_program, build_system)

    @property
    def _directory(self) -> str:
        return self._build_system_rule.build_directory

    @property
    def _make_program(self) -> str:
        make_program = self._provided_make_program
        if self._provided_make_program is None:
            return os.environ.get("MAKE", "make")
        return make_program

    def _pick_directory(
        self, arg: Union[Optional[str], _UNSET] = _UNSET
    ) -> Optional[str]:
        if arg is _UNSET:
            return self._directory
        return arg

    def find_first_existing_make_target(
        self,
        targets: Sequence[str],
        *,
        directory: Union[Optional[str], _UNSET] = _UNSET,
    ) -> Optional[str]:
        for target in targets:
            if self.make_target_exists(target, directory=directory):
                return target
        return None

    def make_target_exists(
        self,
        target: str,
        *,
        directory: Union[Optional[str], _UNSET] = _UNSET,
    ) -> bool:
        return _make_target_exists(
            self._make_program,
            target,
            directory=self._pick_directory(directory),
        )

    def run_first_existing_target_if_any(
        self,
        build_context: "BuildContext",
        targets: Sequence[str],
        *make_args: str,
        enable_parallelization: bool = True,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
    ) -> bool:
        target = self.find_first_existing_make_target(targets, directory=directory)
        if target is None:
            return False

        self.run_make(
            build_context,
            target,
            *make_args,
            enable_parallelization=enable_parallelization,
            directory=directory,
            env_mod=env_mod,
        )
        return True

    def run_make(
        self,
        build_context: "BuildContext",
        *make_args: str,
        enable_parallelization: bool = True,
        directory: Union[Optional[str], _UNSET] = _UNSET,
        env_mod: Optional[EnvironmentModification] = None,
    ) -> None:
        limit = (
            build_context.parallelization_limit(support_zero_as_unlimited=True)
            if enable_parallelization
            else 1
        )
        extra_make_args = [f"-j{limit}"] if limit else ["-j"]
        run_build_system_command(
            self._make_program,
            *extra_make_args,
            *make_args,
            cwd=self._pick_directory(directory),
            env_mod=env_mod,
        )


def debputy_build_system(
    # For future self: Before you get ideas about making manifest_keyword accept a list,
    # remember it has consequences for shadowing_build_systems_when_active.
    manifest_keyword: str,
    provider: Type[BSR],
    *,
    expected_debputy_integration_mode: Optional[
        Container[DebputyIntegrationMode]
    ] = None,
    auto_detection_shadows_build_systems: Optional[
        Union[str, Iterable[str]]
    ] = frozenset(),
    online_reference_documentation: Optional[ParserDocumentation] = None,
    apply_standard_attribute_documentation: bool = False,
    source_format: Optional[Any] = None,
) -> Callable[[Type[BSPF]], Type[BSPF]]:
    if not isinstance(provider, type) or not issubclass(provider, BuildSystemRule):
        raise PluginInitializationError(
            f"The provider parameter for @{debputy_build_system.__name__} must be subclass of"
            f" {BuildSystemRule.__name__}."
        )

    def _constructor_wrapper(
        _rule_used: str,
        *args,
        **kwargs,
    ) -> BSR:
        return provider(*args, **kwargs)

    if isinstance(auto_detection_shadows_build_systems, str):
        shadows = frozenset([auto_detection_shadows_build_systems])
    else:
        shadows = frozenset(auto_detection_shadows_build_systems)

    metadata = BuildSystemManifestRuleMetadata(
        (manifest_keyword,),
        BuildRule,
        _constructor_wrapper,
        expected_debputy_integration_mode=expected_debputy_integration_mode,
        source_format=source_format,
        online_reference_documentation=online_reference_documentation,
        apply_standard_attribute_documentation=apply_standard_attribute_documentation,
        auto_detection_shadow_build_systems=shadows,
        build_system_impl=provider,
    )

    def _decorator_impl(pf_cls: Type[BSPF]) -> Type[BSPF]:
        if isinstance(pf_cls, type) and issubclass(pf_cls, BuildSystemRule):
            raise PluginInitializationError(
                f"The @{debputy_build_system.__name__} annotation goes on the TypedDict that defines the parsed"
                f" variant of the manifest definition. Not the build system implementation class."
            )

        # TODO: In python3.12 we can check more than just `is_typeddict`. In python3.11, woe is us and
        #  is_typeddict is the only thing that reliably works (cpython#103699)
        if not is_typeddict(pf_cls):
            raise PluginInitializationError(
                f"Expected annotated class to be a subclass of {BuildRuleParsedFormat.__name__},"
                f" but got {pf_cls.__name__} instead"
            )

        setattr(pf_cls, _DEBPUTY_DISPATCH_METADATA_ATTR_NAME, metadata)
        return pf_cls

    return _decorator_impl



import re
from typing import (
    Union,
    Sequence,
    Optional,
    Iterable,
    List,
    Mapping,
    TYPE_CHECKING,
    cast,
)

from debputy.filesystem_scan import VirtualPathBase
from debputy.linting.lint_util import LintState
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_completer,
    lsp_semantic_tokens_full,
    SEMANTIC_TOKEN_TYPES_IDS,
    SecondaryLanguage,
    LanguageDispatchRule,
)
from debputy.lsp.quickfixes import (
    propose_remove_range_quick_fix,
    propose_correct_text_quick_fix,
)
from debputy.lsp.text_util import (
    SemanticTokensState,
)
from debputy.lsprotocol.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
    SemanticTokensParams,
    SemanticTokens,
    SemanticTokenTypes,
    Position,
    CompletionItemKind,
    CompletionItemLabelDetails,
)

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types

try:
    from debputy.lsp.vendoring._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
        START_POSITION,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
except ImportError:
    pass


_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/patches/series",
    "patches/series",
    [
        SecondaryLanguage("patches/series", filename_based_lookup=True),
    ],
)


lsp_standard_handler(_DISPATCH_RULE, types.TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)

_RE_LINE_COMMENT = re.compile(r"^\s*(#(?:.*\S)?)\s*$")
_RE_PATCH_LINE = re.compile(
    r"""
    ^  \s* (?P<patch_name> \S+ ) \s*
       (?: (?P<options> [^#\s]+ ) \s* )?
       (?: (?P<comment> \# (?:.*\S)? ) \s* )?
""",
    re.VERBOSE,
)
_RE_UNNECESSARY_LEADING_PREFIX = re.compile(r"(?:(?:[.]{1,2})?/+)+")
_RE_UNNECESSARY_SLASHES = re.compile("//+")


def _all_patch_files(
    debian_patches: VirtualPathBase,
) -> Iterable[VirtualPathBase]:
    if not debian_patches.is_dir:
        return

    for patch_file in debian_patches.all_paths():
        if patch_file.is_dir or patch_file.path in (
            "debian/patches/series",
            "./debian/patches/series",
        ):
            continue

        if patch_file.name.endswith("~"):
            continue
        if patch_file.name.startswith((".#", "#", ".")):
            continue
        parent = patch_file.parent_dir
        if (
            parent is not None
            and parent.path in ("debian/patches", "./debian/patches")
            and patch_file.name.endswith(".series")
        ):
            continue
        yield patch_file


def _listed_patches(
    lines: List[str],
) -> Iterable[str]:
    for line in lines:
        m = _RE_PATCH_LINE.match(line)
        if m is None:
            continue
        filename = m.group(1)
        if filename.startswith("#"):
            continue
        filename = _RE_UNNECESSARY_LEADING_PREFIX.sub("", filename, count=1)
        filename = _RE_UNNECESSARY_SLASHES.sub("/", filename)
        if not filename:
            continue
        yield filename


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_patches_series(lint_state: LintState) -> None:
    source_root = lint_state.source_root
    if source_root is None:
        return None

    dpatches = source_root.lookup("debian/patches/")
    if dpatches is None or not dpatches.is_dir:
        return None

    used_patches = set()
    all_patches = {pf.path for pf in _all_patch_files(dpatches)}

    for line_no, line in enumerate(lint_state.lines):
        m = _RE_PATCH_LINE.match(line)
        if not m:
            continue
        groups = m.groupdict()
        orig_filename = groups["patch_name"]
        filename = orig_filename
        patch_start_col, patch_end_col = m.span("patch_name")
        orig_filename_start_col = patch_start_col
        if filename.startswith("#"):
            continue
        if filename.startswith(("../", "./", "/")):
            sm = _RE_UNNECESSARY_LEADING_PREFIX.match(filename)
            assert sm is not None
            slash_start, slash_end = sm.span(0)
            orig_filename_start_col = slash_end
            prefix = filename[:orig_filename_start_col]
            filename = filename[orig_filename_start_col:]
            slash_range = TERange(
                TEPosition(
                    line_no,
                    patch_start_col + slash_start,
                ),
                TEPosition(
                    line_no,
                    patch_start_col + slash_end,
                ),
            )
            skip_use_check = False
            if ".." in prefix:
                diagnostic_title = f'Disallowed prefix "{prefix}"'
                severity = cast("LintSeverity", "error")
                skip_use_check = True
            else:
                diagnostic_title = f'Unnecessary prefix "{prefix}"'
                severity = cast("LintSeverity", "warning")
            lint_state.emit_diagnostic(
                slash_range,
                diagnostic_title,
                severity,
                "debputy",
                quickfixes=[
                    propose_remove_range_quick_fix(
                        proposed_title=f'Remove prefix "{prefix}"'
                    )
                ],
            )
            if skip_use_check:
                continue
        if "//" in filename:
            for usm in _RE_UNNECESSARY_SLASHES.finditer(filename):
                start_col, end_cold = usm.span()
                slash_range = TERange(
                    TEPosition(
                        line_no,
                        orig_filename_start_col + start_col,
                    ),
                    TEPosition(
                        line_no,
                        orig_filename_start_col + end_cold,
                    ),
                )
                lint_state.emit_diagnostic(
                    slash_range,
                    "Unnecessary slashes",
                    "warning",
                    "debputy",
                    quickfixes=[propose_correct_text_quick_fix("/")],
                )
            filename = _RE_UNNECESSARY_SLASHES.sub("/", filename)

        patch_name_range = TERange(
            TEPosition(
                line_no,
                patch_start_col,
            ),
            TEPosition(
                line_no,
                patch_end_col,
            ),
        )
        if not filename.lower().endswith((".diff", ".patch")):
            lint_state.emit_diagnostic(
                patch_name_range,
                f'Patch not using ".patch" or ".diff" as extension: "{filename}"',
                "pedantic",
                "debputy",
            )
        patch_path = f"{dpatches.path}/{filename}"
        if patch_path not in all_patches:
            lint_state.emit_diagnostic(
                patch_name_range,
                f'Non-existing patch "{filename}"',
                "error",
                "debputy",
            )
        elif patch_path in used_patches:
            lint_state.emit_diagnostic(
                patch_name_range,
                f'Duplicate patch: "{filename}"',
                "error",
                "debputy",
            )
        else:
            used_patches.add(patch_path)

    unused_patches = all_patches - used_patches
    for unused_patch in sorted(unused_patches):
        patch_name = unused_patch[len(dpatches.path) + 1 :]
        line_count = len(lint_state.lines)
        file_range = TERange(
            TEPosition(
                0,
                0,
            ),
            TEPosition(
                line_count,
                len(lint_state.lines[-1]) if line_count else 0,
            ),
        )
        lint_state.emit_diagnostic(
            file_range,
            f'Unused patch: "{patch_name}"',
            "warning",
            "debputy",
        )


@lsp_completer(_DISPATCH_RULE)
def _debian_patches_series_completions(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> Optional[Union[CompletionList, Sequence[CompletionItem]]]:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lint_state = ls.lint_state(doc)
    source_root = lint_state.source_root
    dpatches = source_root.lookup("debian/patches") if source_root is not None else None
    if dpatches is None:
        return None
    lines = doc.lines
    position = doc.position_codec.position_from_client_units(lines, params.position)
    line = lines[position.line]
    if line.startswith("#"):
        return None
    try:
        line.rindex(" #", 0, position.character)
        return None  # In an end of line comment
    except ValueError:
        pass
    already_used = set(_listed_patches(lines))
    # `debian/patches + "/"`
    dpatches_dir_len = len(dpatches.path) + 1
    all_patch_files_gen = (
        p.path[dpatches_dir_len:] for p in _all_patch_files(dpatches)
    )
    return [
        CompletionItem(
            p,
            kind=CompletionItemKind.File,
            insert_text=f"{p}\n",
            label_details=CompletionItemLabelDetails(
                description=f"debian/patches/{p}",
            ),
        )
        for p in all_patch_files_gen
        if p not in already_used
    ]


@lsp_semantic_tokens_full(_DISPATCH_RULE)
async def _debian_patches_semantic_tokens_full(
    ls: "DebputyLanguageServer",
    request: SemanticTokensParams,
) -> Optional[SemanticTokens]:
    doc = ls.workspace.get_text_document(request.text_document.uri)
    lines = doc.lines
    position_codec = doc.position_codec

    tokens: List[int] = []
    string_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.String.value]
    comment_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Comment.value]
    options_token_code = SEMANTIC_TOKEN_TYPES_IDS[SemanticTokenTypes.Keyword.value]
    sem_token_state = SemanticTokensState(
        ls,
        doc,
        lines,
        tokens,
    )

    async for line_no, line in ls.slow_iter(enumerate(lines)):
        if line.isspace():
            continue
        m = _RE_LINE_COMMENT.match(line)
        if m:
            start_col, end_col = m.span(1)
            start_pos = position_codec.position_to_client_units(
                sem_token_state.lines,
                Position(
                    line_no,
                    start_col,
                ),
            )
            sem_token_state.emit_token(
                start_pos,
                position_codec.client_num_units(line[start_col:end_col]),
                comment_token_code,
            )
            continue
        m = _RE_PATCH_LINE.match(line)
        if not m:
            continue
        groups = m.groupdict()
        _emit_group(
            line_no,
            string_token_code,
            sem_token_state,
            "patch_name",
            groups,
            m,
        )
        _emit_group(
            line_no,
            options_token_code,
            sem_token_state,
            "options",
            groups,
            m,
        )
        _emit_group(
            line_no,
            comment_token_code,
            sem_token_state,
            "comment",
            groups,
            m,
        )

    return SemanticTokens(tokens)


def _emit_group(
    line_no: int,
    token_code: int,
    sem_token_state: SemanticTokensState,
    group_name: str,
    groups: Mapping[str, str],
    match: re.Match,
) -> None:
    value = groups.get(group_name)
    if not value:
        return None
    patch_start_col = match.start(group_name)
    position_codec = sem_token_state.doc.position_codec
    patch_start_pos = position_codec.position_to_client_units(
        sem_token_state.lines,
        Position(
            line_no,
            patch_start_col,
        ),
    )
    sem_token_state.emit_token(
        patch_start_pos,
        position_codec.client_num_units(value),
        token_code,
    )



import dataclasses
import json
import os
import subprocess
import textwrap
from typing import (
    NotRequired,
    TypedDict,
    Self,
    cast,
    Dict,
    Mapping,
    Sequence,
    MutableMapping,
    Iterable,
    Container,
    List,
    Tuple,
    Union,
    Optional,
    TYPE_CHECKING,
    Literal,
)

from debian.debian_support import Version

from debputy._manifest_constants import MK_BUILDS
from debputy.manifest_parser.base_types import (
    BuildEnvironmentDefinition,
    DebputyParsedContentStandardConditional,
    FileSystemExactMatchRule,
)
from debputy.manifest_parser.exceptions import (
    ManifestParseException,
    ManifestInvalidUserDataException,
)
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.util import AttributePath
from debputy.plugin.api import reference_documentation
from debputy.plugin.api.impl import (
    DebputyPluginInitializerProvider,
)
from debputy.plugin.api.parser_tables import OPARSER_MANIFEST_ROOT
from debputy.plugin.api.spec import (
    documented_attr,
    INTEGRATION_MODE_FULL,
    only_integrations,
    VirtualPath,
)
from debputy.plugin.api.std_docs import docs_from
from debputy.plugin.debputy.to_be_api_types import (
    BuildRule,
    StepBasedBuildSystemRule,
    OptionalInstallDirectly,
    BuildSystemCharacteristics,
    OptionalBuildDirectory,
    OptionalInSourceBuild,
    MakefileSupport,
    BuildRuleParsedFormat,
    debputy_build_system,
    CleanHelper,
    NinjaBuildSupport,
)
from debputy.types import EnvironmentModification
from debputy.util import (
    _warn,
    run_build_system_command,
    _error,
    PerlConfigVars,
    resolve_perl_config,
    generated_content_dir,
    manifest_format_doc,
)

if TYPE_CHECKING:
    from debputy.build_support.build_context import BuildContext
    from debputy.highlevel_manifest import HighLevelManifest


PERL_CMD = "perl"


def register_build_system_rules(api: DebputyPluginInitializerProvider) -> None:
    register_build_keywords(api)
    register_build_rules(api)


def register_build_keywords(api: DebputyPluginInitializerProvider) -> None:

    api.pluggable_manifest_rule(
        OPARSER_MANIFEST_ROOT,
        "build-environments",
        List[NamedEnvironmentSourceFormat],
        _parse_build_environments,
        expected_debputy_integration_mode=only_integrations(INTEGRATION_MODE_FULL),
        inline_reference_documentation=reference_documentation(
            title="Build Environments (`build-environments`)",
            description=textwrap.dedent(
                """\
                Define named environments to set the environment for any build commands that needs
                a non-default environment.

                The environment definitions can be used to tweak the environment variables used by the
                build commands. An example:

                    build-environments:
                      - name: custom-env
                        set:
                          ENV_VAR: foo
                          ANOTHER_ENV_VAR: bar
                    builds:
                      - autoconf:
                          environment: custom-env

                The environment definition has multiple attributes for setting environment variables
                which determines when the definition is applied. The resulting environment is the
                result of the following order of operations.

                  1. The environment `debputy` received from its parent process.
                  2. Apply all the variable definitions from `set` (if the attribute is present)
                  3. Apply all computed variables (such as variables from `dpkg-buildflags`).
                  4. Apply all the variable definitions from `override` (if the attribute is present)
                  5. Remove all variables listed in `unset` (if the attribute is present).

                Accordingly, both `override` and `unset` will overrule any computed variables while
                `set` will be overruled by any computed variables.

                Note that these variables are not available via manifest substitution (they are only
                visible to build commands). They are only available to build commands.
            """
            ),
            attributes=[
                documented_attr(
                    "name",
                    textwrap.dedent(
                        """\
                        The name of the environment

                        The name is used to reference the environment from build rules.
                    """
                    ),
                ),
                documented_attr(
                    "set",
                    textwrap.dedent(
                        """\
                        A mapping of environment variables to be set.

                        Note these environment variables are set before computed variables (such
                        as `dpkg-buildflags`) are provided. They can affect the content of the
                        computed variables, but they cannot overrule them. If you need to overrule
                        a computed variable, please use `override` instead.
                """
                    ),
                ),
                documented_attr(
                    "override",
                    textwrap.dedent(
                        """\
                        A mapping of environment variables to set.

                        Similar to `set`, but it can overrule computed variables like those from
                        `dpkg-buildflags`.
                """
                    ),
                ),
                documented_attr(
                    "unset",
                    textwrap.dedent(
                        """\
                        A list of environment variables to unset.

                        Any environment variable named here will be unset. No warnings or errors
                        will be raised if a given variable was not set.
                    """
                    ),
                ),
            ],
            reference_documentation_url=manifest_format_doc(
                f"build-environment-build-environment"
            ),
        ),
    )
    api.pluggable_manifest_rule(
        OPARSER_MANIFEST_ROOT,
        "default-build-environment",
        EnvironmentSourceFormat,
        _parse_default_environment,
        expected_debputy_integration_mode=only_integrations(INTEGRATION_MODE_FULL),
        inline_reference_documentation=reference_documentation(
            title="Default Build Environment (`default-build-environment`)",
            description=textwrap.dedent(
                """\
                Define the environment variables used in all build commands that uses the default
                environment.

                The environment definition can be used to tweak the environment variables used by the
                build commands. An example:
                
                    default-build-environment:
                      set:
                        ENV_VAR: foo
                        ANOTHER_ENV_VAR: bar

                The environment definition has multiple attributes for setting environment variables
                which determines when the definition is applied. The resulting environment is the
                result of the following order of operations.
                
                  1. The environment `debputy` received from its parent process.
                  2. Apply all the variable definitions from `set` (if the attribute is present)
                  3. Apply all computed variables (such as variables from `dpkg-buildflags`).
                  4. Apply all the variable definitions from `override` (if the attribute is present)
                  5. Remove all variables listed in `unset` (if the attribute is present).

                Accordingly, both `override` and `unset` will overrule any computed variables while
                `set` will be overruled by any computed variables.

                Note that these variables are not available via manifest substitution (they are only
                visible to build commands). They are only available to build commands.
            """
            ),
            attributes=[
                documented_attr(
                    "set",
                    textwrap.dedent(
                        """\
                        A mapping of environment variables to be set.

                        Note these environment variables are set before computed variables (such
                        as `dpkg-buildflags`) are provided. They can affect the content of the
                        computed variables, but they cannot overrule them. If you need to overrule
                        a computed variable, please use `override` instead.
                """
                    ),
                ),
                documented_attr(
                    "override",
                    textwrap.dedent(
                        """\
                        A mapping of environment variables to set.

                        Similar to `set`, but it can overrule computed variables like those from
                        `dpkg-buildflags`.
                """
                    ),
                ),
                documented_attr(
                    "unset",
                    textwrap.dedent(
                        """\
                        A list of environment variables to unset.

                        Any environment variable named here will be unset. No warnings or errors
                        will be raised if a given variable was not set.
                    """
                    ),
                ),
            ],
            reference_documentation_url=manifest_format_doc(
                "build-environment-build-environment"
            ),
        ),
    )
    api.pluggable_manifest_rule(
        OPARSER_MANIFEST_ROOT,
        MK_BUILDS,
        List[BuildRule],
        _handle_build_rules,
        expected_debputy_integration_mode=only_integrations(INTEGRATION_MODE_FULL),
        inline_reference_documentation=reference_documentation(
            title=f"Build rules (`{MK_BUILDS}`)",
            description=textwrap.dedent(
                """\
                Define how to build the upstream part of the package. Usually this is done via "build systems",
                which also defines the clean rules.

                A simple example is:

                ```yaml
                    builds:
                      - autoconf:
                          configure-args:
                            - "--enable-foo"
                            - "--without=bar"
                ```

                Multiple builds are supported out of box. Here an example of how the build rules
                for `libpam-krb5` might look.

                ```yaml
                    builds:
                      - autoconf:
                          for: libpam-krb5
                          install-directly-to-package: true
                          configure-args:
                            - "--enable-reduced-depends"
                            - "--with-krb5-include=/usr/include/mit-krb5"
                            - "--with-krb5-lib=/usr/lib/{{DEB_HOST_MULTIARCH}}/mit-krb5"
                            - "--with-kadm-client-include=/usr/include/mit-krb5"
                            - "--with-kadm-client-lib=/usr/lib/{{DEB_HOST_MULTIARCH}}/mit-krb5"
                      - autoconf:
                          for: libpam-heimdal
                          install-directly-to-package: true
                          configure-args:
                            - "--enable-reduced-depends"
                            - "--with-krb5-include=/usr/include/heimdal"
                            - "--with-krb5-lib=/usr/lib/{{DEB_HOST_MULTIARCH}}/heimdal"
                            - "--with-kadm-client-include=/usr/include/heimdal"
                            - "--with-kadm-client-lib=/usr/lib/{{DEB_HOST_MULTIARCH}}/heimdal"
                ```
            """
            ),
        ),
    )


def register_build_rules(api: DebputyPluginInitializerProvider) -> None:
    api.register_build_system(ParsedAutoconfBuildRuleDefinition)
    api.register_build_system(ParsedMakeBuildRuleDefinition)

    api.register_build_system(ParsedPerlBuildBuildRuleDefinition)
    api.register_build_system(ParsedPerlMakeMakerBuildRuleDefinition)
    api.register_build_system(ParsedDebhelperBuildRuleDefinition)

    api.register_build_system(ParsedCMakeBuildRuleDefinition)
    api.register_build_system(ParsedMesonBuildRuleDefinition)

    api.register_build_system(ParsedQmakeBuildRuleDefinition)
    api.register_build_system(ParsedQmake6BuildRuleDefinition)


class EnvironmentSourceFormat(TypedDict):
    set: NotRequired[Dict[str, str]]
    override: NotRequired[Dict[str, str]]
    unset: NotRequired[List[str]]


class NamedEnvironmentSourceFormat(EnvironmentSourceFormat):
    name: str


_READ_ONLY_ENV_VARS = {
    "DEB_CHECK_COMMAND": None,
    "DEB_SIGN_KEYID": None,
    "DEB_SIGN_KEYFILE": None,
    "DEB_BUILD_OPTIONS": "DEB_BUILD_MAINT_OPTIONS",
    "DEB_BUILD_PROFILES": None,
    "DEB_RULES_REQUIRES_ROOT": None,
    "DEB_GAIN_ROOT_COMMAND": None,
    "DH_EXTRA_ADDONS": None,
    "DH_NO_ACT": None,
}


def _check_variables(
    env_vars: Iterable[str],
    attribute_path: AttributePath,
) -> None:
    for env_var in env_vars:
        if env_var not in _READ_ONLY_ENV_VARS:
            continue
        alt = _READ_ONLY_ENV_VARS.get(env_var)
        var_path = attribute_path[env_var].path_key_lc
        if alt is None:
            raise ManifestParseException(
                f"The variable {env_var} cannot be modified by the manifest. This restriction is generally"
                f" because the build should not touch those variables or changing them have no effect"
                f" (since the consumer will not see the change). The problematic definition was {var_path}"
            )
        else:
            raise ManifestParseException(
                f"The variable {env_var} cannot be modified by the manifest. This restriction is generally"
                f" because the build should not touch those variables or changing them have no effect"
                f" (since the consumer will not see the change). Depending on what you are trying to"
                f' accomplish, the variable "{alt}" might be a suitable alternative.'
                f" The problematic definition was {var_path}"
            )


def _no_overlap(
    lhs: Iterable[Union[str, Tuple[int, str]]],
    rhs: Container[str],
    lhs_key: str,
    rhs_key: str,
    redundant_key: str,
    attribute_path: AttributePath,
) -> None:
    for kt in lhs:
        if isinstance(kt, tuple):
            lhs_path_key, var = kt
        else:
            lhs_path_key = var = kt
        if var not in rhs:
            continue
        lhs_path = attribute_path[lhs_key][lhs_path_key].path_key_lc
        rhs_path = attribute_path[rhs_key][var].path_key_lc
        r_path = lhs_path if redundant_key == rhs_key else rhs_path
        raise ManifestParseException(
            f"The environment variable {var} was declared in {lhs_path} and {rhs_path}."
            f" Due to how the variables are applied, the definition in {r_path} is redundant"
            f" and can effectively be removed. Please review the manifest and remove one of"
            f" the two definitions."
        )


@dataclasses.dataclass(slots=True, frozen=True)
class ManifestProvidedBuildEnvironment(BuildEnvironmentDefinition):

    name: str
    is_default: bool
    attribute_path: AttributePath
    parser_context: ParserContextData

    set_vars: Mapping[str, str]
    override_vars: Mapping[str, str]
    unset_vars: Sequence[str]

    @classmethod
    def from_environment_definition(
        cls,
        env: EnvironmentSourceFormat,
        attribute_path: AttributePath,
        parser_context: ParserContextData,
        is_default: bool = False,
    ) -> Self:
        reference_name: Optional[str]
        if is_default:
            name = "default-env"
            reference_name = None
        else:
            named_env = cast("NamedEnvironmentSourceFormat", env)
            name = named_env["name"]
            reference_name = name

        set_vars = env.get("set", {})
        override_vars = env.get("override", {})
        unset_vars = env.get("unset", [])
        _check_variables(set_vars, attribute_path["set"])
        _check_variables(override_vars, attribute_path["override"])
        _check_variables(unset_vars, attribute_path["unset"])

        if not set_vars and not override_vars and not unset_vars:
            raise ManifestParseException(
                f"The environment definition {attribute_path.path_key_lc} was empty. Please provide"
                " some content or delete the definition."
            )

        _no_overlap(
            enumerate(unset_vars),
            set_vars,
            "unset",
            "set",
            "set",
            attribute_path,
        )
        _no_overlap(
            enumerate(unset_vars),
            override_vars,
            "unset",
            "override",
            "override",
            attribute_path,
        )
        _no_overlap(
            override_vars,
            set_vars,
            "override",
            "set",
            "set",
            attribute_path,
        )

        r = cls(
            name,
            is_default,
            attribute_path,
            parser_context,
            set_vars,
            override_vars,
            unset_vars,
        )
        parser_context._register_build_environment(
            reference_name,
            r,
            attribute_path,
            is_default,
        )

        return r

    def update_env(self, env: MutableMapping[str, str]) -> None:
        if set_vars := self.set_vars:
            env.update(set_vars)
        dpkg_env = self.dpkg_buildflags_env(env, self.attribute_path.path_key_lc)
        self.log_computed_env(f"dpkg-buildflags [{self.name}]", dpkg_env)
        if overlapping_env := dpkg_env.keys() & set_vars.keys():
            for var in overlapping_env:
                key_lc = self.attribute_path["set"][var].path_key_lc
                _warn(
                    f'The variable "{var}" defined at {key_lc} is shadowed by a computed variable.'
                    f" If the manifest definition is more important, please define it via `override` rather than"
                    f" `set`."
                )
        env.update(dpkg_env)
        if override_vars := self.override_vars:
            env.update(override_vars)
        if unset_vars := self.unset_vars:
            for var in unset_vars:
                try:
                    del env[var]
                except KeyError:
                    pass


_MAKE_DEFAULT_TOOLS = [
    ("CC", "gcc"),
    ("CXX", "g++"),
    ("PKG_CONFIG", "pkg-config"),
]


class MakefileBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = ("_make_support", "_build_target", "_install_target", "_directory")

    def __init__(
        self,
        attributes: "ParsedMakeBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(attributes, attribute_path, parser_context)
        directory = attributes.get("directory")
        if directory is not None:
            self._directory = directory.match_rule.path
        else:
            self._directory = None
        self._make_support = MakefileSupport.from_build_system(self)
        self._build_target = attributes.get("build_target")
        self._test_target = attributes.get("test_target")
        self._install_target = attributes.get("install_target")

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return any(p in source_root for p in ("Makefile", "makefile", "GNUmakefile"))

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="not-supported",
        )

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        # No configure step
        pass

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        extra_vars = []
        build_target = self._build_target
        if build_target is not None:
            extra_vars.append(build_target)
        if context.is_cross_compiling:
            for envvar, tool in _MAKE_DEFAULT_TOOLS:
                cross_tool = os.environ.get(envvar)
                if cross_tool is None:
                    cross_tool = context.cross_tool(tool)
                extra_vars.append(f"{envvar}={cross_tool}")
        self._make_support.run_make(
            context,
            *extra_vars,
            "INSTALL=install --strip-program=true",
            directory=self._directory,
        )

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._run_make_maybe_explicit_target(
            context,
            self._test_target,
            ["test", "check"],
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        self._run_make_maybe_explicit_target(
            context,
            self._install_target,
            ["install"],
            f"DESTDIR={dest_dir}",
            "AM_UPDATE_INFO_DIR=no",
            "INSTALL=install --strip-program=true",
        )

    def _run_make_maybe_explicit_target(
        self,
        context: "BuildContext",
        provided_target: Optional[str],
        fallback_targets: Sequence[str],
        *make_args: str,
    ) -> None:
        make_support = self._make_support
        if provided_target is not None:
            make_support.run_make(
                context,
                provided_target,
                *make_args,
                directory=self._directory,
            )
        else:
            make_support.run_first_existing_target_if_any(
                context,
                fallback_targets,
                *make_args,
                directory=self._directory,
            )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        self._make_support.run_first_existing_target_if_any(
            context,
            ["distclean", "realclean", "clean"],
        )


class PerlBuildBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = "configure_args"

    def __init__(
        self,
        attributes: "ParsedPerlBuildBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(attributes, attribute_path, parser_context)
        self.configure_args = attributes.get("configure_args", [])

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return "Build.PL" in source_root

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="not-supported",
        )

    @staticmethod
    def _perl_cross_build_env(
        context: "BuildContext",
    ) -> Tuple[PerlConfigVars, Optional[EnvironmentModification]]:
        perl_config_data = resolve_perl_config(
            context.dpkg_architecture_variables,
            None,
        )
        if context.is_cross_compiling:
            perl5lib_dir = perl_config_data.cross_inc_dir
            if perl5lib_dir is not None:
                env_perl5lib = os.environ.get("PERL5LIB")
                if env_perl5lib is not None:
                    perl5lib_dir = (
                        perl5lib_dir + perl_config_data.path_sep + env_perl5lib
                    )
                env_mod = EnvironmentModification(
                    replacements=(("PERL5LIB", perl5lib_dir),),
                )
                return perl_config_data, env_mod
        return perl_config_data, None

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        perl_config_data, cross_env_mod = self._perl_cross_build_env(context)
        configure_env = EnvironmentModification(
            replacements=(
                ("PERL_MM_USE_DEFAULT", "1"),
                ("PKG_CONFIG", context.cross_tool("pkg-config")),
            )
        )
        if cross_env_mod is not None:
            configure_env = configure_env.combine(cross_env_mod)

        configure_cmd = [
            PERL_CMD,
            "Build.PL",
            "--installdirs",
            "vendor",
        ]
        cflags = os.environ.get("CFLAGS", "")
        cppflags = os.environ.get("CPPFLAGS", "")
        ldflags = os.environ.get("LDFLAGS", "")

        if cflags != "" or cppflags != "":
            configure_cmd.append("--config")
            combined = f"{cflags} {cppflags}".strip()
            configure_cmd.append(f"optimize={combined}")

        if ldflags != "" or cflags != "" or context.is_cross_compiling:
            configure_cmd.append("--config")
            combined = f"{perl_config_data.ld} {cflags} {ldflags}".strip()
            configure_cmd.append(f"ld={combined}")
        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            configure_cmd.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )
        run_build_system_command(*configure_cmd, env_mod=configure_env)

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        _, cross_env_mod = self._perl_cross_build_env(context)
        run_build_system_command(PERL_CMD, "Build", env_mod=cross_env_mod)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        _, cross_env_mod = self._perl_cross_build_env(context)
        run_build_system_command(
            PERL_CMD,
            "Build",
            "test",
            "--verbose",
            "1",
            env_mod=cross_env_mod,
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        _, cross_env_mod = self._perl_cross_build_env(context)
        run_build_system_command(
            PERL_CMD,
            "Build",
            "install",
            "--destdir",
            dest_dir,
            "--create_packlist",
            "0",
            env_mod=cross_env_mod,
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        _, cross_env_mod = self._perl_cross_build_env(context)
        if os.path.lexists("Build"):
            run_build_system_command(
                PERL_CMD,
                "Build",
                "realclean",
                "--allow_mb_mismatch",
                "1",
                env_mod=cross_env_mod,
            )


class PerlMakeMakerBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = ("configure_args", "_make_support")

    def __init__(
        self,
        attributes: "ParsedPerlBuildBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(attributes, attribute_path, parser_context)
        self.configure_args = attributes.get("configure_args", [])
        self._make_support = MakefileSupport.from_build_system(self)

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return "Makefile.PL" in source_root

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="not-supported",
        )

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        configure_env = EnvironmentModification(
            replacements=(
                ("PERL_MM_USE_DEFAULT", "1"),
                ("PERL_AUTOINSTALL", "--skipdeps"),
                ("PKG_CONFIG", context.cross_tool("pkg-config")),
            )
        )
        perl_args = []
        mm_args = ["INSTALLDIRS=vendor"]
        if "CFLAGS" in os.environ:
            mm_args.append(
                f"OPTIMIZE={os.environ['CFLAGS']} {os.environ['CPPFLAGS']}".rstrip()
            )

        perl_config_data = resolve_perl_config(
            context.dpkg_architecture_variables,
            None,
        )

        if "LDFLAGS" in os.environ:
            mm_args.append(
                f"LD={perl_config_data.ld} {os.environ['CFLAGS']} {os.environ['LDFLAGS']}"
            )

        if context.is_cross_compiling:
            perl5lib_dir = perl_config_data.cross_inc_dir
            if perl5lib_dir is not None:
                perl_args.append(f"-I{perl5lib_dir}")

        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            mm_args.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )
        run_build_system_command(
            PERL_CMD,
            *perl_args,
            "Makefile.PL",
            *mm_args,
            env_mod=configure_env,
        )

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._make_support.run_make(context)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._make_support.run_first_existing_target_if_any(
            context,
            ["check", "test"],
            "TEST_VERBOSE=1",
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        is_mm_makefile = False
        with open("Makefile", "rb") as fd:
            for line in fd:
                if b"generated automatically by MakeMaker" in line:
                    is_mm_makefile = True
                    break

        install_args = [f"DESTDIR={dest_dir}"]

        # Special case for Makefile.PL that uses
        # Module::Build::Compat. PREFIX should not be passed
        # for those; it already installs into /usr by default.
        if is_mm_makefile:
            install_args.append("PREFIX=/usr")

        self._make_support.run_first_existing_target_if_any(
            context,
            ["install"],
            *install_args,
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        self._make_support.run_first_existing_target_if_any(
            context,
            ["distclean", "realclean", "clean"],
        )


class DebhelperBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = ("configure_args", "dh_build_system")

    def __init__(
        self,
        parsed_data: "ParsedDebhelperBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(parsed_data, attribute_path, parser_context)
        self.configure_args = parsed_data.get("configure_args", [])
        self.dh_build_system = parsed_data.get("dh_build_system")

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        try:
            v = subprocess.check_output(
                ["dh_assistant", "which-build-system"],
                # Packages without `debhelper-compat` will trigger an error, which will just be noise
                stderr=subprocess.DEVNULL,
                cwd=source_root.fs_path,
            )
        except subprocess.CalledProcessError:
            return False
        else:
            d = json.loads(v)
            build_system = d.get("build-system")
            return build_system is not None

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="supported-but-not-default",
        )

    def before_first_impl_step(
        self, *, stage: Literal["build", "clean"], **kwargs
    ) -> None:
        dh_build_system = self.dh_build_system
        if dh_build_system is None:
            return
        try:
            subprocess.check_call(
                ["dh_assistant", "which-build-system", f"-S{dh_build_system}"]
            )
        except FileNotFoundError:
            _error(
                "The debhelper build system assumes `dh_assistant` is available (`debhelper (>= 13.5~)`)"
            )
        except subprocess.SubprocessError:
            raise ManifestInvalidUserDataException(
                f'The debhelper build system "{dh_build_system}" does not seem to'
                f" be available according to"
                f" `dh_assistant which-build-system -S{dh_build_system}`"
            ) from None

    def _default_options(self) -> List[str]:
        default_options = []
        if self.dh_build_system is not None:
            default_options.append(f"-S{self.dh_build_system}")
        if self.build_directory is not None:
            default_options.append(f"-B{self.build_directory}")

        return default_options

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        if (
            os.path.lexists("configure.ac") or os.path.lexists("configure.in")
        ) and not os.path.lexists("debian/autoreconf.before"):
            run_build_system_command("dh_update_autotools_config")
            run_build_system_command("dh_autoreconf")

        default_options = self._default_options()
        configure_args = default_options.copy()
        if self.configure_args:
            configure_args.append("--")
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            configure_args.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )
        run_build_system_command("dh_auto_configure", *configure_args)

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        default_options = self._default_options()
        run_build_system_command("dh_auto_build", *default_options)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        default_options = self._default_options()
        run_build_system_command("dh_auto_test", *default_options)

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        default_options = self._default_options()
        run_build_system_command(
            "dh_auto_install",
            *default_options,
            f"--destdir={dest_dir}",
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        default_options = self._default_options()
        run_build_system_command("dh_auto_clean", *default_options)
        # The "global" clean logic takes care of `dh_autoreconf_clean` and `dh_clean`


class AutoconfBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = ("configure_args", "_make_support")

    def __init__(
        self,
        parsed_data: "ParsedAutoconfBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(parsed_data, attribute_path, parser_context)
        configure_args = [a for a in parsed_data.get("configure_args", [])]
        self.configure_args = configure_args
        self._make_support = MakefileSupport.from_build_system(self)

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="supported-and-default",
        )

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        if "configure.ac" in source_root:
            return True
        configure_in = source_root.get("configure.in")
        if configure_in is not None and configure_in.is_file:
            with configure_in.open(byte_io=True, buffering=4096) as fd:
                for no, line in enumerate(fd):
                    if no > 100:
                        break
                    if b"AC_INIT" in line or b"AC_PREREQ" in line:
                        return True
        configure = source_root.get("configure")
        if configure is None or not configure.is_executable or not configure.is_file:
            return False
        with configure.open(byte_io=True, buffering=4096) as fd:
            for no, line in enumerate(fd):
                if no > 10:
                    break
                if b"GNU Autoconf" in line:
                    return True
        return False

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        if (
            os.path.lexists("configure.ac") or os.path.lexists("configure.in")
        ) and not os.path.lexists("debian/autoreconf.before"):
            run_build_system_command("dh_update_autotools_config")
            run_build_system_command("dh_autoreconf")

        dpkg_architecture_variables = context.dpkg_architecture_variables
        multi_arch = dpkg_architecture_variables.current_host_multiarch
        silent_rules = (
            "--enable-silent-rules"
            if context.is_terse_build
            else "--disable-silent-rules"
        )

        configure_args = [
            f"--build={dpkg_architecture_variables['DEB_BUILD_GNU_TYPE']}",
            "--prefix=/usr",
            "--includedir=${prefix}/include",
            "--mandir=${prefix}/share/man",
            "--infodir=${prefix}/share/info",
            "--sysconfdir=/etc",
            "--localstatedir=/var",
            "--disable-option-checking",
            silent_rules,
            f"--libdir=${{prefix}}/{multi_arch}",
            "--runstatedir=/run",
            "--disable-maintainer-mode",
            "--disable-dependency-tracking",
        ]
        if dpkg_architecture_variables.is_cross_compiling:
            configure_args.append(
                f"--host={dpkg_architecture_variables['DEB_HOST_GNU_TYPE']}"
            )
        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            configure_args.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )
        self.ensure_build_dir_exists()
        configure_script = self.relative_from_builddir_to_source("configure")
        with self.dump_logs_on_error("config.log"):
            run_build_system_command(
                configure_script,
                *configure_args,
                cwd=self.build_directory,
            )

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._make_support.run_make(context)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        limit = context.parallelization_limit(support_zero_as_unlimited=True)
        testsuite_flags = [f"-j{limit}"] if limit else ["-j"]

        if not context.is_terse_build:
            testsuite_flags.append("--verbose")
        self._make_support.run_first_existing_target_if_any(
            context,
            # Order is deliberately inverse compared to debhelper (#924052)
            ["check", "test"],
            f"TESTSUITEFLAGS={' '.join(testsuite_flags)}",
            "VERBOSE=1",
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        enable_parallelization = not os.path.lexists(self.build_dir_path("libtool"))
        self._make_support.run_first_existing_target_if_any(
            context,
            ["install"],
            f"DESTDIR={dest_dir}",
            "AM_UPDATE_INFO_DIR=no",
            enable_parallelization=enable_parallelization,
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        if self.out_of_source_build:
            return
        self._make_support.run_first_existing_target_if_any(
            context,
            ["distclean", "realclean", "clean"],
        )
        # The "global" clean logic takes care of `dh_autoreconf_clean` and `dh_clean`


class CMakeBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = (
        "configure_args",
        "target_build_system",
        "_make_support",
        "_ninja_support",
    )

    def __init__(
        self,
        parsed_data: "ParsedCMakeBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(parsed_data, attribute_path, parser_context)
        configure_args = [a for a in parsed_data.get("configure_args", [])]
        self.configure_args = configure_args
        self.target_build_system: Literal["make", "ninja"] = parsed_data.get(
            "target_build_system", "make"
        )
        self._make_support = MakefileSupport.from_build_system(self)
        self._ninja_support = NinjaBuildSupport.from_build_system(self)

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="required",
        )

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return "CMakeLists.txt" in source_root

    @staticmethod
    def _default_cmake_env(
        build_context: "BuildContext",
    ) -> EnvironmentModification:
        replacements = {}
        if "DEB_PYTHON_INSTALL_LAYOUT" not in os.environ:
            replacements["DEB_PYTHON_INSTALL_LAYOUT"] = "deb"
        if "PKG_CONFIG" not in os.environ:
            pkg_config = build_context.cross_tool("pkg-config")
            replacements["PKG_CONFIG"] = f"/usr/bin/{pkg_config}"
        return EnvironmentModification(
            replacements=tuple((k, v) for k, v in replacements.items())
        )

    @classmethod
    def cmake_generator(cls, target_build_system: Literal["make", "ninja"]) -> str:
        cmake_generators = {
            "make": "Unix Makefiles",
            "ninja": "Ninja",
        }
        return cmake_generators[target_build_system]

    @staticmethod
    def _compiler_and_cross_flags(
        context: "BuildContext",
        cmake_flags: List[str],
    ) -> None:

        if "CC" in os.environ:
            cmake_flags.append(f"-DCMAKE_C_COMPILER={os.environ['CC']}")
        elif context.is_cross_compiling:
            cmake_flags.append(f"-DCMAKE_C_COMPILER={context.cross_tool('gcc')}")

        if "CXX" in os.environ:
            cmake_flags.append(f"-DCMAKE_CXX_COMPILER={os.environ['CXX']}")
        elif context.is_cross_compiling:
            cmake_flags.append(f"-DCMAKE_CXX_COMPILER={context.cross_tool('g++')}")

        if context.is_cross_compiling:
            deb_host2cmake_system = {
                "linux": "Linux",
                "kfreebsd": "kFreeBSD",
                "hurd": "GNU",
            }

            gnu_cpu2system_processor = {
                "arm": "armv7l",
                "misp64el": "mips64",
                "powerpc64le": "ppc64le",
            }
            dpkg_architecture_variables = context.dpkg_architecture_variables

            try:
                system_name = deb_host2cmake_system[
                    dpkg_architecture_variables["DEB_HOST_ARCH_OS"]
                ]
            except KeyError as e:
                name = e.args[0]
                _error(
                    f"Cannot cross-compile via cmake: Missing CMAKE_SYSTEM_NAME for the DEB_HOST_ARCH_OS {name}"
                )

            gnu_cpu = dpkg_architecture_variables["DEB_HOST_GNU_CPU"]
            system_processor = gnu_cpu2system_processor.get(gnu_cpu, gnu_cpu)

            cmake_flags.append(f"-DCMAKE_SYSTEM_NAME={system_name}")
            cmake_flags.append(f"-DCMAKE_SYSTEM_PROCESSOR={system_processor}")

            pkg_config = context.cross_tool("pkg-config")
            # Historical uses. Current versions of cmake uses the env variable instead.
            cmake_flags.append(f"-DPKG_CONFIG_EXECUTABLE=/usr/bin/{pkg_config}")
            cmake_flags.append(f"-DPKGCONFIG_EXECUTABLE=/usr/bin/{pkg_config}")
            cmake_flags.append(
                f"-DQMAKE_EXECUTABLE=/usr/bin/{context.cross_tool('qmake')}"
            )

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        cmake_flags = [
            "-DCMAKE_INSTALL_PREFIX=/usr",
            "-DCMAKE_BUILD_TYPE=None",
            "-DCMAKE_INSTALL_SYSCONFDIR=/etc",
            "-DCMAKE_INSTALL_LOCALSTATEDIR=/var",
            "-DCMAKE_EXPORT_NO_PACKAGE_REGISTRY=ON",
            "-DCMAKE_FIND_USE_PACKAGE_REGISTRY=OFF",
            "-DCMAKE_FIND_PACKAGE_NO_PACKAGE_REGISTRY=ON",
            "-DFETCHCONTENT_FULLY_DISCONNECTED=ON",
            "-DCMAKE_INSTALL_RUNSTATEDIR=/run",
            "-DCMAKE_SKIP_INSTALL_ALL_DEPENDENCY=ON",
            "-DCMAKE_BUILD_RPATH_USE_ORIGIN=ON",
            f"-DCMAKE_INSTALL_LIBDIR=lib/{context.dpkg_architecture_variables['DEB_HOST_MULTIARCH']}",
            f"-G{self.cmake_generator(self.target_build_system)}",
        ]
        if not context.is_terse_build:
            cmake_flags.append("-DCMAKE_VERBOSE_MAKEFILE=ON")
        if not context.should_run_tests:
            cmake_flags.append("-DBUILD_TESTING:BOOL=OFF")

        self._compiler_and_cross_flags(context, cmake_flags)

        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            cmake_flags.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )

        env_mod = self._default_cmake_env(context)
        if "CPPFLAGS" in os.environ:
            # CMake doesn't respect CPPFLAGS, see #653916.
            cppflags = os.environ["CPPFLAGS"]
            cflags = f"{os.environ.get('CFLAGS', '')} {cppflags}".lstrip()
            cxxflags = f"{os.environ.get('CXXFLAGS', '')} {cppflags}".lstrip()
            env_mod = env_mod.combine(
                # The debhelper build system never showed this delta, so people might find it annoying.
                EnvironmentModification(
                    replacements=(
                        ("CFLAGS", cflags),
                        ("CXXFLAGS", cxxflags),
                    )
                )
            )
        if "ASMFLAGS" not in os.environ and "ASFLAGS" in os.environ:
            env_mod = env_mod.combine(
                # The debhelper build system never showed this delta, so people might find it annoying.
                EnvironmentModification(
                    replacements=(("ASMFLAGS", os.environ["ASFLAGS"]),),
                )
            )
        self.ensure_build_dir_exists()
        source_dir_from_build_dir = self.relative_from_builddir_to_source()

        with self.dump_logs_on_error(
            "CMakeCache.txt",
            "CMakeFiles/CMakeOutput.log",
            "CMakeFiles/CMakeError.log",
        ):
            run_build_system_command(
                "cmake",
                *cmake_flags,
                source_dir_from_build_dir,
                cwd=self.build_directory,
                env_mod=env_mod,
            )

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        if self.target_build_system == "make":
            make_flags = []
            if not context.is_terse_build:
                make_flags.append("VERBOSE=1")
            self._make_support.run_make(context, *make_flags)
        else:
            self._ninja_support.run_ninja_build(context)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        env_mod = EnvironmentModification(
            replacements=(("CTEST_OUTPUT_ON_FAILURE", "1"),),
        )
        if self.target_build_system == "make":
            # Unlike make, CTest does not have "unlimited parallel" setting (-j implies
            # -j1). Therefore, we do not set "allow zero as unlimited" here.
            make_flags = [f"ARGS+=-j{context.parallelization_limit()}"]
            if not context.is_terse_build:
                make_flags.append("ARGS+=--verbose")
            self._make_support.run_first_existing_target_if_any(
                context,
                ["check", "test"],
                *make_flags,
                env_mod=env_mod,
            )
        else:
            self._ninja_support.run_ninja_test(context, env_mod=env_mod)

        limit = context.parallelization_limit(support_zero_as_unlimited=True)
        testsuite_flags = [f"-j{limit}"] if limit else ["-j"]

        if not context.is_terse_build:
            testsuite_flags.append("--verbose")
        self._make_support.run_first_existing_target_if_any(
            context,
            # Order is deliberately inverse compared to debhelper (#924052)
            ["check", "test"],
            f"TESTSUITEFLAGS={' '.join(testsuite_flags)}",
            "VERBOSE=1",
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        env_mod = EnvironmentModification(
            replacements=(
                ("LC_ALL", "C.UTF-8"),
                ("DESTDIR", dest_dir),
            )
        ).combine(self._default_cmake_env(context))
        run_build_system_command(
            "cmake",
            "--install",
            self.build_directory,
            env_mod=env_mod,
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        if self.out_of_source_build:
            return
        if self.target_build_system == "make":
            # Keep it here in case we change the `required` "out of source" to "supported-default"
            self._make_support.run_first_existing_target_if_any(
                context,
                ["distclean", "realclean", "clean"],
            )
        else:
            self._ninja_support.run_ninja_clean(context)


class MesonBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = (
        "configure_args",
        "_ninja_support",
    )

    def __init__(
        self,
        parsed_data: "ParsedMesonBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(parsed_data, attribute_path, parser_context)
        configure_args = [a for a in parsed_data.get("configure_args", [])]
        self.configure_args = configure_args
        self._ninja_support = NinjaBuildSupport.from_build_system(self)

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="required",
        )

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return "meson.build" in source_root

    @staticmethod
    def _default_meson_env() -> EnvironmentModification:
        replacements = {
            "LC_ALL": "C.UTF-8",
        }
        if "DEB_PYTHON_INSTALL_LAYOUT" not in os.environ:
            replacements["DEB_PYTHON_INSTALL_LAYOUT"] = "deb"
        return EnvironmentModification(
            replacements=tuple((k, v) for k, v in replacements.items())
        )

    @classmethod
    def cmake_generator(cls, target_build_system: Literal["make", "ninja"]) -> str:
        cmake_generators = {
            "make": "Unix Makefiles",
            "ninja": "Ninja",
        }
        return cmake_generators[target_build_system]

    @staticmethod
    def _cross_flags(
        context: "BuildContext",
        meson_flags: List[str],
    ) -> None:
        if not context.is_cross_compiling:
            return
        # Needs a cross-file http://mesonbuild.com/Cross-compilation.html
        cross_files_dir = os.path.abspath(
            generated_content_dir(
                subdir_key="meson-cross-files",
            )
        )
        host_arch = context.dpkg_architecture_variables.current_host_arch
        cross_file = os.path.join(cross_files_dir, f"meson-cross-file-{host_arch}.conf")
        if not os.path.isfile(cross_file):
            env = os.environ
            if env.get("LC_ALL") != "C.UTF-8":
                env = dict(env)
                env["LC_ALL"] = "C.UTF-8"
            else:
                env = None
            subprocess.check_call(
                [
                    "/usr/share/meson/debcrossgen",
                    f"--arch={host_arch}",
                    f"-o{cross_file}",
                ],
                stdout=subprocess.DEVNULL,
                env=env,
            )

        meson_flags.append("--cross-file")
        meson_flags.append(cross_file)

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        meson_version = Version(
            subprocess.check_output(
                ["meson", "--version"],
                encoding="utf-8",
            ).strip()
        )
        dpkg_architecture_variables = context.dpkg_architecture_variables

        meson_flags = [
            "--wrap-mode=nodownload",
            "--buildtype=plain",
            "--prefix=/usr",
            "--sysconfdir=/etc",
            "--localstatedir=/var",
            f"--libdir=lib/{dpkg_architecture_variables.current_host_multiarch}",
            "--auto-features=enabled",
        ]
        if meson_version >= Version("1.2.0"):
            # There was a behavior change in Meson 1.2.0: previously
            # byte-compilation wasn't supported, but since 1.2.0 it is on by
            # default. We can only use this option to turn it off in versions
            # where the option exists.
            meson_flags.append("-Dpython.bytecompile=-1")

        self._cross_flags(context, meson_flags)

        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            meson_flags.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )

        env_mod = self._default_meson_env()

        self.ensure_build_dir_exists()
        source_dir_from_build_dir = self.relative_from_builddir_to_source()

        with self.dump_logs_on_error("meson-logs/meson-log.txt"):
            run_build_system_command(
                "meson",
                "setup",
                source_dir_from_build_dir,
                *meson_flags,
                cwd=self.build_directory,
                env_mod=env_mod,
            )

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._ninja_support.run_ninja_build(context)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        env_mod = EnvironmentModification(
            replacements=(("MESON_TESTTHREADS", f"{context.parallelization_limit()}"),),
        ).combine(self._default_meson_env())
        meson_args = []
        if not context.is_terse_build:
            meson_args.append("--verbose")
        with self.dump_logs_on_error("meson-logs/testlog.txt"):
            run_build_system_command(
                "meson",
                "test",
                *meson_args,
                env_mod=env_mod,
                cwd=self.build_directory,
            )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        run_build_system_command(
            "meson",
            "install",
            "--destdir",
            dest_dir,
            cwd=self.build_directory,
            env_mod=self._default_meson_env(),
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        # `debputy` will handle all the cleanup for us by virtue of "out of source build"
        assert self.out_of_source_build


def _add_qmake_flag(options: List[str], envvar: str, *, include_cppflags: bool) -> None:
    value = os.environ.get(envvar)
    if value is None:
        return
    if include_cppflags:
        cppflags = os.environ.get("CPPFLAGS")
        if cppflags:
            value = f"{value} {cppflags}"

    options.append(f"QMAKE_{envvar}_RELEASE={value}")
    options.append(f"QMAKE_{envvar}_DEBUG={value}")


class ParsedGenericQmakeBuildRuleDefinition(
    OptionalInstallDirectly,
    OptionalInSourceBuild,
    OptionalBuildDirectory,
):
    configure_args: NotRequired[List[str]]


class AbstractQmakeBuildSystemRule(StepBasedBuildSystemRule):

    __slots__ = ("configure_args", "_make_support")

    def __init__(
        self,
        parsed_data: "ParsedGenericQmakeBuildRuleDefinition",
        attribute_path: AttributePath,
        parser_context: Union[ParserContextData, "HighLevelManifest"],
    ) -> None:
        super().__init__(parsed_data, attribute_path, parser_context)
        configure_args = [a for a in parsed_data.get("configure_args", [])]
        self.configure_args = configure_args
        self._make_support = MakefileSupport.from_build_system(self)

    @classmethod
    def characteristics(cls) -> BuildSystemCharacteristics:
        return BuildSystemCharacteristics(
            out_of_source_builds="supported-and-default",
        )

    @classmethod
    def auto_detect_build_system(
        cls,
        source_root: VirtualPath,
        *args,
        **kwargs,
    ) -> bool:
        return any(p.name.endswith(".pro") for p in source_root.iterdir)

    @classmethod
    def os_mkspec_mapping(cls) -> Mapping[str, str]:
        return {
            "linux": "linux-g++",
            "kfreebsd": "gnukfreebsd-g++",
            "hurd": "hurd-g++",
        }

    def qmake_command(self) -> str:
        raise NotImplementedError

    def configure_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:

        configure_args = [
            "-makefile",
        ]
        qmake_cmd = context.cross_tool(self.qmake_command())

        if context.is_cross_compiling:
            host_os = context.dpkg_architecture_variables["DEB_HOST_ARCH_OS"]
            os2mkspec = self.os_mkspec_mapping()
            try:
                spec = os2mkspec[host_os]
            except KeyError:
                _error(
                    f'Sorry, `debputy` cannot cross build this package for "{host_os}".'
                    f' Missing a "DEB OS -> qmake -spec <VALUE>" mapping.'
                )
            configure_args.append("-spec")
            configure_args.append(spec)

        _add_qmake_flag(configure_args, "CFLAGS", include_cppflags=True)
        _add_qmake_flag(configure_args, "CXXFLAGS", include_cppflags=True)
        _add_qmake_flag(configure_args, "LDFLAGS", include_cppflags=False)

        configure_args.append("QMAKE_STRIP=:")
        configure_args.append("PREFIX=/usr")

        if self.configure_args:
            substitution = self.substitution
            attr_path = self.attribute_path["configure_args"]
            configure_args.extend(
                substitution.substitute(v, attr_path[i].path)
                for i, v in enumerate(self.configure_args)
            )

        self.ensure_build_dir_exists()
        if not self.out_of_source_build:
            configure_args.append(self.relative_from_builddir_to_source())

        with self.dump_logs_on_error("config.log"):
            run_build_system_command(
                qmake_cmd,
                *configure_args,
                cwd=self.build_directory,
            )

    def build_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        self._make_support.run_make(context)

    def test_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        **kwargs,
    ) -> None:
        limit = context.parallelization_limit(support_zero_as_unlimited=True)
        testsuite_flags = [f"-j{limit}"] if limit else ["-j"]

        if not context.is_terse_build:
            testsuite_flags.append("--verbose")
        self._make_support.run_first_existing_target_if_any(
            context,
            # Order is deliberately inverse compared to debhelper (#924052)
            ["check", "test"],
            f"TESTSUITEFLAGS={' '.join(testsuite_flags)}",
            "VERBOSE=1",
        )

    def install_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        dest_dir: str,
        **kwargs,
    ) -> None:
        enable_parallelization = not os.path.lexists(self.build_dir_path("libtool"))
        self._make_support.run_first_existing_target_if_any(
            context,
            ["install"],
            f"DESTDIR={dest_dir}",
            "AM_UPDATE_INFO_DIR=no",
            enable_parallelization=enable_parallelization,
        )

    def clean_impl(
        self,
        context: "BuildContext",
        manifest: "HighLevelManifest",
        clean_helper: "CleanHelper",
        **kwargs,
    ) -> None:
        if self.out_of_source_build:
            return
        self._make_support.run_first_existing_target_if_any(
            context,
            ["distclean", "realclean", "clean"],
        )


class QmakeBuildSystemRule(AbstractQmakeBuildSystemRule):

    def qmake_command(self) -> str:
        return "qmake"


class Qmake6BuildSystemRule(AbstractQmakeBuildSystemRule):

    def qmake_command(self) -> str:
        return "qmake6"


@debputy_build_system(
    "make",
    MakefileBuildSystemRule,
    auto_detection_shadows_build_systems="debhelper",
    online_reference_documentation=reference_documentation(
        title="Make Build System",
        description=textwrap.dedent(
            """\
        Run a plain `make` file with nothing else.

        This build system will attempt to use `make` to leverage instructions
        in a makefile (such as, `Makefile` or `GNUMakefile`).

        By default, the makefile build system assumes it should use "in-source"
        build semantics. If needed be, an explicit `build-directory` can be
        provided if the `Makefile` is not in the source folder but instead in
        some other directory.
        """
        ),
        attributes=[
            documented_attr(
                "directory",
                textwrap.dedent(
                    """\
                The directory from which to run make if it is not the source root

                This works like using `make -C DIRECTORY ...` (or `cd DIRECTORY && make ...`).
                """
                ),
            ),
            documented_attr(
                "build_target",
                textwrap.dedent(
                    """\
                The target name to use for the "build" step.

                If omitted, `make` will be run without any explicit target leaving it to decide
                the default.
                """
                ),
            ),
            documented_attr(
                "test_target",
                textwrap.dedent(
                    """\
                The target name to use for the "test" step.

                If omitted, `make check` or `make test` will be used if it looks like `make`
                will accept one of those targets. Otherwise, the step will be skipped.
                """
                ),
            ),
            documented_attr(
                "install_target",
                textwrap.dedent(
                    """\
                The target name to use for the "install" step.

                If omitted, `make install` will be used if it looks like `make` will accept that target.
                Otherwise, the step will be skipped.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedMakeBuildRuleDefinition(
    OptionalInstallDirectly,
):
    directory: NotRequired[FileSystemExactMatchRule]
    build_target: NotRequired[str]
    test_target: NotRequired[str]
    install_target: NotRequired[str]


@debputy_build_system(
    "autoconf",
    AutoconfBuildSystemRule,
    auto_detection_shadows_build_systems=["debhelper", "make"],
    online_reference_documentation=reference_documentation(
        title="Autoconf Build System (`autoconf`)",
        description=textwrap.dedent(
            """\
        Run an autoconf-based build system as the upstream build system.

        This build rule will attempt to use autoreconf to update the `configure`
        script before running the `configure` script if needed. Otherwise, it
        follows the classic `./configure && make && make install` pattern.

        The build rule uses "out of source" builds by default since it is easier
        and more reliable for clean and makes it easier to support multiple
        builds (that is, two or more build systems for the same source package).
        This is in contract to `debhelper`, which defaults to "in source" builds
        for `autoconf`. If you need that behavior, please set
        `perform-in-source-build: true`.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `configure` script.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalInSourceBuild,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedAutoconfBuildRuleDefinition(
    OptionalInstallDirectly,
    OptionalInSourceBuild,
    OptionalBuildDirectory,
):
    configure_args: NotRequired[List[str]]


@debputy_build_system(
    "cmake",
    CMakeBuildSystemRule,
    auto_detection_shadows_build_systems=["debhelper", "make"],
    online_reference_documentation=reference_documentation(
        title="CMake Build System (`cmake`)",
        description=textwrap.dedent(
            """\
        Run an cmake-based build system as the upstream build system.

        The build rule uses "out of source" builds.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `cmake` command.
                """
                ),
            ),
            documented_attr(
                "target_build_system",
                textwrap.dedent(
                    """\
                    Which build system should `cmake` should delegate the build system to.
                    In `cmake` terms, this is a Generator (the `-G` option).

                    Supported options are:
                     * `make` - GNU Make
                     * `ninja` - Ninja
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedCMakeBuildRuleDefinition(
    OptionalInstallDirectly,
    OptionalBuildDirectory,
):
    configure_args: NotRequired[List[str]]
    target_build_system: Literal["make", "ninja"]


@debputy_build_system(
    "meson",
    MesonBuildSystemRule,
    auto_detection_shadows_build_systems=["debhelper", "make"],
    online_reference_documentation=reference_documentation(
        title="Meson Build System (`meson`)",
        description=textwrap.dedent(
            """\
        Run an meson-based build system as the upstream build system.

        The build rule uses "out of source" builds.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `meson` command.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedMesonBuildRuleDefinition(
    OptionalInstallDirectly,
    OptionalBuildDirectory,
):
    configure_args: NotRequired[List[str]]


@debputy_build_system(
    "perl-build",
    PerlBuildBuildSystemRule,
    auto_detection_shadows_build_systems=[
        "debhelper",
        "make",
        "perl-makemaker",
    ],
    online_reference_documentation=reference_documentation(
        title='Perl "Build.PL" Build System (`perl-build`)',
        description=textwrap.dedent(
            """\
        Build using the `Build.PL` Build system used by some Perl packages.

        This build rule will attempt to use the `Build.PL` script to build the
        upstream code.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `Build.PL` script.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedPerlBuildBuildRuleDefinition(
    OptionalInstallDirectly,
):
    configure_args: NotRequired[List[str]]


@debputy_build_system(
    "debhelper",
    DebhelperBuildSystemRule,
    online_reference_documentation=reference_documentation(
        title="Debhelper Build System (`debhelper`)",
        description=textwrap.dedent(
            """\
        Delegate to a debhelper provided build system

        This build rule will attempt to use the `dh_auto_*` tools to build the
        upstream code. By default, `dh_auto_*` will use auto-detection to determine
        which build system they will use. This can be overridden by the
        `dh-build-system` attribute.
        """
        ),
        attributes=[
            documented_attr(
                "dh_build_system",
                textwrap.dedent(
                    """\
                    Which debhelper build system to use. This attribute is passed to
                    the `dh_auto_*` commands as the `-S` parameter, so any value valid
                    for that will be accepted.

                    Note that many debhelper build systems require extra build
                    dependencies before they can be used. Please consult the documentation
                    of the relevant debhelper build system for details.
                """
                ),
            ),
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to underlying configuration command
                     (via `dh_auto_configure -- <configure-args`).
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedDebhelperBuildRuleDefinition(
    OptionalInstallDirectly,
    OptionalBuildDirectory,
):
    configure_args: NotRequired[List[str]]
    dh_build_system: NotRequired[str]


@debputy_build_system(
    "perl-makemaker",
    PerlMakeMakerBuildSystemRule,
    auto_detection_shadows_build_systems=[
        "debhelper",
        "make",
    ],
    online_reference_documentation=reference_documentation(
        title='Perl "MakeMaker" Build System (`perl-makemaker`)',
        description=textwrap.dedent(
            """\
        Build using the "MakeMaker" Build system used by some Perl packages.

        This build rule will attempt to use the `Makefile.PL` script to build the
        upstream code.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `Makefile.PL` script.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedPerlMakeMakerBuildRuleDefinition(
    OptionalInstallDirectly,
):
    configure_args: NotRequired[List[str]]


@debputy_build_system(
    "qmake",
    QmakeBuildSystemRule,
    auto_detection_shadows_build_systems=[
        "debhelper",
        "make",
        # Open question, should this shadow "qmake6" and later?
    ],
    online_reference_documentation=reference_documentation(
        title='QT "qmake" Build System (`qmake`)',
        description=textwrap.dedent(
            """\
        Build using the "qmake" by QT.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `qmake` command.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalInSourceBuild,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedQmakeBuildRuleDefinition(ParsedGenericQmakeBuildRuleDefinition):
    pass


@debputy_build_system(
    "qmake6",
    Qmake6BuildSystemRule,
    auto_detection_shadows_build_systems=[
        "debhelper",
        "make",
    ],
    online_reference_documentation=reference_documentation(
        title='QT "qmake6" Build System (`qmake6`)',
        description=textwrap.dedent(
            """\
        Build using the "qmake6" from the `qmake6` package.  This is like the `qmake` system
        but is specifically for QT6.
        """
        ),
        attributes=[
            documented_attr(
                "configure_args",
                textwrap.dedent(
                    """\
                    Arguments to be passed to the `qmake6` command.
                """
                ),
            ),
            *docs_from(
                DebputyParsedContentStandardConditional,
                OptionalInstallDirectly,
                OptionalInSourceBuild,
                OptionalBuildDirectory,
                BuildRuleParsedFormat,
            ),
        ],
    ),
)
class ParsedQmake6BuildRuleDefinition(ParsedGenericQmakeBuildRuleDefinition):
    pass


def _parse_default_environment(
    _name: str,
    parsed_data: EnvironmentSourceFormat,
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> ManifestProvidedBuildEnvironment:
    return ManifestProvidedBuildEnvironment.from_environment_definition(
        parsed_data,
        attribute_path,
        parser_context,
        is_default=True,
    )


def _parse_build_environments(
    _name: str,
    parsed_data: List[NamedEnvironmentSourceFormat],
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> List[ManifestProvidedBuildEnvironment]:
    return [
        ManifestProvidedBuildEnvironment.from_environment_definition(
            value,
            attribute_path[idx],
            parser_context,
            is_default=False,
        )
        for idx, value in enumerate(parsed_data)
    ]


def _handle_build_rules(
    _name: str,
    parsed_data: List[BuildRule],
    _attribute_path: AttributePath,
    _parser_context: ParserContextData,
) -> List[BuildRule]:
    return parsed_data



from debputy.plugin.api.impl_types import (
    KNOWN_PACKAGING_FILE_CATEGORY_DESCRIPTIONS,
    KNOWN_PACKAGING_FILE_CONFIG_FEATURE_DESCRIPTION,
)

REFERENCE_DATA_TABLE = {
    "config-features": KNOWN_PACKAGING_FILE_CONFIG_FEATURE_DESCRIPTION,
    "file-categories": KNOWN_PACKAGING_FILE_CATEGORY_DESCRIPTIONS,
}
